import pandas as pd
import sqlalchemy as sa
import sqlalchemy.dialects.mysql as mysql

from ibis.sql.alchemy import (unary, fixed_arity, infix_op,
                              _variance_reduction)
import ibis.common as com
import ibis.expr.types as ir
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.sql.alchemy as alch


_operation_registry = alch._operation_registry.copy()

# NOTE: window functions are available from MySQL 8 and MariaDB 10.2
_operation_registry.update(alch._window_functions)


def _substr(t, expr):
    f = sa.func.substr

    arg, start, length = expr.op().args

    sa_arg = t.translate(arg)
    sa_start = t.translate(start)

    if length is None:
        return f(sa_arg, sa_start + 1)
    else:
        sa_length = t.translate(length)
        return f(sa_arg, sa_start + 1, sa_length)


def _string_find(t, expr):
    arg, substr, start, _ = expr.op().args

    if start is not None:
        raise NotImplementedError

    sa_arg = t.translate(arg)
    sa_substr = t.translate(substr)

    return sa.func.locate(sa_arg, sa_substr) - 1


def _capitalize(t, expr):
    arg, = expr.op().args
    sa_arg = t.translate(arg)
    return sa.func.concat(
        sa.func.ucase(
            sa.func.left(sa_arg, 1)
        ),
        sa.func.substring(sa_arg, 2)
    )


def _extract(fmt):
    def translator(t, expr):
        arg, = expr.op().args
        sa_arg = t.translate(arg)
        return sa.extract(fmt, sa_arg)
    return translator


_truncate_formats = {
    's': '%Y-%m-%d %H:%i:%s',
    'm': '%Y-%m-%d %H:%i:00',
    'h': '%Y-%m-%d %H:00:00',
    'D': '%Y-%m-%d',
    # 'W': 'week',
    'M': '%Y-%m-01',
    'Y': '%Y-01-01'
}


def _truncate(t, expr):
    arg, unit = expr.op().args
    sa_arg = t.translate(arg)
    try:
        fmt = _truncate_formats[unit]
    except KeyError:
        raise com.UnsupportedOperationError(
            'Unsupported truncate unit {}'.format(unit)
        )
    return sa.func.date_format(sa_arg, fmt)


def _cast(t, expr):
    arg, typ = expr.op().args

    sa_arg = t.translate(arg)
    sa_type = t.get_sqla_type(typ)

    # specialize going from an integer type to a timestamp
    if isinstance(arg.type(), dt.Integer) and isinstance(sa_type, sa.DateTime):
        return sa.func.timezone('UTC', sa.func.to_timestamp(sa_arg))

    if arg.type().equals(dt.binary) and typ.equals(dt.string):
        return sa.func.encode(sa_arg, 'escape')

    if typ.equals(dt.binary):
        #  decode yields a column of memoryview which is annoying to deal with
        # in pandas. CAST(expr AS BYTEA) is correct and returns byte strings.
        return sa.cast(sa_arg, sa.Binary())

    return sa.cast(sa_arg, sa_type)


def _log(t, expr):
    arg, base = expr.op().args
    sa_arg = t.translate(arg)
    sa_base = t.translate(base)
    return sa.func.log(sa_base, sa_arg)


def _identical_to(t, expr):
    left, right = args = expr.op().args
    if left.equals(right):
        return True
    else:
        left, right = map(t.translate, args)
        return left.op('<=>')(right)


def _round(t, expr):
    arg, digits = expr.op().args
    sa_arg = t.translate(arg)

    if digits is None:
        sa_digits = 0
    else:
        sa_digits = t.translate(digits)

    return sa.func.round(sa_arg, sa_digits)


def _floor_divide(t, expr):
    left, right = map(t.translate, expr.op().args)
    return sa.func.floor(left / right)


def _string_join(t, expr):
    sep, elements = expr.op().args
    return sa.func.concat_ws(t.translate(sep), *map(t.translate, elements))


def _interval_from_integer(t, expr):
    arg, unit = expr.op().args
    if unit in {'ms', 'ns'}:
        raise com.UnsupportedOperationError(
            'MySQL does not allow operation '
            'with INTERVAL offset {}'.format(unit)
        )

    sa_arg = t.translate(arg)
    return sa.text('INTERVAL {} {}'.format(sa_arg, expr.resolution.upper()))


def _timestamp_diff(t, expr):
    left, right = expr.op().args
    sa_left = t.translate(left)
    sa_right = t.translate(right)
    return sa.func.timestampdiff(sa.text('SECOND'), sa_right, sa_left)


def _literal(t, expr):
    if isinstance(expr, ir.IntervalValue):
        if expr.type().unit in {'ms', 'ns'}:
            raise com.UnsupportedOperationError(
                'MySQL does not allow operation '
                'with INTERVAL offset {}'.format(expr.type().unit)
            )
        return sa.text('INTERVAL {} {}'.format(expr.op().value,
                                               expr.resolution.upper()))
    else:
        value = expr.op().value
        if isinstance(value, pd.Timestamp):
            value = value.to_pydatetime()
        return sa.literal(value)


_operation_registry.update({
    ir.Literal: _literal,

    # strings
    ops.Substring: _substr,
    ops.StringFind: _string_find,
    ops.Capitalize: _capitalize,
    ops.RegexSearch: infix_op('REGEXP'),

    # math
    ops.Log: _log,
    ops.Log2: unary(sa.func.log2),
    ops.Log10: unary(sa.func.log10),
    ops.Round: _round,

    # dates and times
    ops.Date: unary(sa.func.date),
    ops.DateAdd: infix_op('+'),
    ops.DateSub: infix_op('-'),
    ops.DateDiff: fixed_arity(sa.func.datediff, 2),
    ops.TimestampAdd: infix_op('+'),
    ops.TimestampSub: infix_op('-'),
    ops.TimestampDiff: _timestamp_diff,
    ops.DateTruncate: _truncate,
    ops.TimestampTruncate: _truncate,
    ops.IntervalFromInteger: _interval_from_integer,
    ops.Strftime: fixed_arity(sa.func.date_format, 2),
    ops.ExtractYear: _extract('year'),
    ops.ExtractMonth: _extract('month'),
    ops.ExtractDay: _extract('day'),
    ops.ExtractHour: _extract('hour'),
    ops.ExtractMinute: _extract('minute'),
    ops.ExtractSecond: _extract('second'),
    ops.ExtractMillisecond: _extract('millisecond'),

    # reductions
    ops.Variance: _variance_reduction('var'),
    ops.StandardDev: _variance_reduction('stddev'),

    ops.IdenticalTo: _identical_to
})


def add_operation(op, translation_func):
    _operation_registry[op] = translation_func


class MySQLExprTranslator(alch.AlchemyExprTranslator):

    _registry = _operation_registry
    _rewrites = alch.AlchemyExprTranslator._rewrites.copy()
    _type_map = alch.AlchemyExprTranslator._type_map.copy()
    _type_map.update({
        dt.Int8: mysql.TINYINT,
        dt.Double: mysql.DOUBLE,
        dt.Float: mysql.FLOAT,
    })


rewrites = MySQLExprTranslator.rewrites
compiles = MySQLExprTranslator.compiles


class MySQLDialect(alch.AlchemyDialect):

    translator = MySQLExprTranslator


dialect = MySQLDialect



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlalchemy as sa

import toolz

from ibis.sql.alchemy import unary, varargs, fixed_arity, _variance_reduction
import ibis.sql.alchemy as alch
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
import ibis.common as com


_operation_registry = alch._operation_registry.copy()


def _cast(t, expr):
    # It's not all fun and games with SQLite

    op = expr.op()
    arg, target_type = op.args
    sa_arg = t.translate(arg)
    sa_type = t.get_sqla_type(target_type)

    if isinstance(target_type, dt.Timestamp):
        if isinstance(arg, ir.IntegerValue):
            return sa.func.datetime(sa_arg, 'unixepoch')
        elif isinstance(arg, ir.StringValue):
            return sa.func.strftime('%Y-%m-%d %H:%M:%f', sa_arg)
        raise com.UnsupportedOperationError(type(arg))

    if isinstance(target_type, dt.Date):
        if isinstance(arg, ir.IntegerValue):
            return sa.func.date(sa.func.datetime(sa_arg, 'unixepoch'))
        elif isinstance(arg, ir.StringValue):
            return sa.func.date(sa_arg)
        raise com.UnsupportedOperationError(type(arg))

    if isinstance(arg, ir.CategoryValue) and target_type == 'int32':
        return sa_arg
    else:
        return sa.cast(sa_arg, sa_type)


def _substr(t, expr):
    f = sa.func.substr

    arg, start, length = expr.op().args

    sa_arg = t.translate(arg)
    sa_start = t.translate(start)

    if length is None:
        return f(sa_arg, sa_start + 1)
    else:
        sa_length = t.translate(length)
        return f(sa_arg, sa_start + 1, sa_length)


def _string_right(t, expr):
    f = sa.func.substr

    arg, length = expr.op().args

    sa_arg = t.translate(arg)
    sa_length = t.translate(length)

    return f(sa_arg, -sa_length, sa_length)


def _string_find(t, expr):
    arg, substr, start, _ = expr.op().args

    if start is not None:
        raise NotImplementedError

    sa_arg = t.translate(arg)
    sa_substr = t.translate(substr)

    f = sa.func.instr
    return f(sa_arg, sa_substr) - 1


def _infix_op(infix_sym):
    def formatter(t, expr):
        op = expr.op()
        left, right = op.args

        left_arg = t.translate(left)
        right_arg = t.translate(right)
        return left_arg.op(infix_sym)(right_arg)

    return formatter


def _strftime(t, expr):
    arg, format = expr.op().args
    sa_arg = t.translate(arg)
    sa_format = t.translate(format)
    return sa.func.strftime(sa_format, sa_arg)


def _strftime_int(fmt):
    def translator(t, expr):
        arg, = expr.op().args
        sa_arg = t.translate(arg)
        return sa.cast(sa.func.strftime(fmt, sa_arg), sa.INTEGER)
    return translator


_truncate_modifiers = {
    'Y': 'start of year',
    'M': 'start of month',
    'D': 'start of day',
    'W': 'weekday 1',
}


def _truncate(func):
    def translator(t, expr):
        arg, unit = expr.op().args
        sa_arg = t.translate(arg)
        try:
            modifier = _truncate_modifiers[unit]
        except KeyError:
            raise com.UnsupportedOperationError(
                'Unsupported truncate unit {!r}'.format(unit)
            )
        return func(sa_arg, modifier)

    return translator


def _now(t, expr):
    return sa.func.datetime('now')


def _millisecond(t, expr):
    arg, = expr.op().args
    sa_arg = t.translate(arg)
    fractional_second = sa.func.strftime('%f', sa_arg)
    return (fractional_second * 1000) % 1000


def _identical_to(t, expr):
    left, right = args = expr.op().args
    if left.equals(right):
        return True
    else:
        left, right = map(t.translate, args)
        return sa.func.coalesce(
            (left.is_(None) & right.is_(None)) | (left == right),
            False
        )


def _log(t, expr):
    arg, base = expr.op().args
    sa_arg = t.translate(arg)
    if base is None:
        return sa.func._ibis_sqlite_ln(sa_arg)
    return sa.func._ibis_sqlite_log(sa_arg, t.translate(base))


def _repeat(t, expr):
    arg, times = map(t.translate, expr.op().args)
    f = sa.func
    return f.replace(
        f.substr(
            f.quote(
                f.zeroblob((times + 1) / 2)
            ),
            3,
            times
        ),
        '0',
        arg
    )


def _generic_pad(arg, length, pad):
    f = sa.func
    arg_length = f.length(arg)
    pad_length = f.length(pad)
    number_of_zero_bytes = (
        (length - arg_length - 1 + pad_length) / pad_length + 1) / 2
    return f.substr(
        f.replace(
            f.replace(
                f.substr(f.quote(f.zeroblob(number_of_zero_bytes)), 3),
                "'",
                ''
            ),
            '0',
            pad
        ),
        1,
        length - f.length(arg)
    )


def _lpad(t, expr):
    arg, length, pad = map(t.translate, expr.op().args)
    return _generic_pad(arg, length, pad) + arg


def _rpad(t, expr):
    arg, length, pad = map(t.translate, expr.op().args)
    return arg + _generic_pad(arg, length, pad)


_operation_registry.update({
    ops.Cast: _cast,

    ops.Substring: _substr,
    ops.StrRight: _string_right,

    ops.StringFind: _string_find,

    ops.Least: varargs(sa.func.min),
    ops.Greatest: varargs(sa.func.max),
    ops.IfNull: fixed_arity(sa.func.ifnull, 2),

    ops.DateTruncate: _truncate(sa.func.date),
    ops.TimestampTruncate: _truncate(sa.func.datetime),
    ops.Strftime: _strftime,
    ops.ExtractYear: _strftime_int('%Y'),
    ops.ExtractMonth: _strftime_int('%m'),
    ops.ExtractDay: _strftime_int('%d'),
    ops.ExtractHour: _strftime_int('%H'),
    ops.ExtractMinute: _strftime_int('%M'),
    ops.ExtractSecond: _strftime_int('%S'),
    ops.ExtractMillisecond: _millisecond,
    ops.TimestampNow: _now,
    ops.IdenticalTo: _identical_to,

    ops.RegexSearch: fixed_arity(sa.func._ibis_sqlite_regex_search, 2),
    ops.RegexReplace: fixed_arity(sa.func._ibis_sqlite_regex_replace, 3),
    ops.RegexExtract: fixed_arity(sa.func._ibis_sqlite_regex_extract, 3),

    ops.LPad: _lpad,
    ops.RPad: _rpad,
    ops.Repeat: _repeat,
    ops.Reverse: unary(sa.func._ibis_sqlite_reverse),

    ops.StringAscii: unary(sa.func._ibis_sqlite_string_ascii),
    ops.Capitalize: unary(sa.func._ibis_sqlite_capitalize),
    ops.Translate: fixed_arity(sa.func._ibis_sqlite_translate, 3),

    ops.Sqrt: unary(sa.func._ibis_sqlite_sqrt),
    ops.Power: fixed_arity(sa.func._ibis_sqlite_power, 2),
    ops.Exp: unary(sa.func._ibis_sqlite_exp),
    ops.Ln: unary(sa.func._ibis_sqlite_ln),
    ops.Log: _log,
    ops.Log10: unary(sa.func._ibis_sqlite_log10),
    ops.Log2: unary(sa.func._ibis_sqlite_log2),
    ops.Floor: unary(sa.func._ibis_sqlite_floor),
    ops.Ceil: unary(sa.func._ibis_sqlite_ceil),
    ops.Sign: unary(sa.func._ibis_sqlite_sign),
    ops.FloorDivide: fixed_arity(sa.func._ibis_sqlite_floordiv, 2),

    ops.Variance: _variance_reduction('_ibis_sqlite_var'),
    ops.StandardDev: toolz.compose(
        sa.func._ibis_sqlite_sqrt,
        _variance_reduction('_ibis_sqlite_var')
    ),
})


def add_operation(op, translation_func):
    _operation_registry[op] = translation_func


class SQLiteExprTranslator(alch.AlchemyExprTranslator):

    _registry = _operation_registry
    _rewrites = alch.AlchemyExprTranslator._rewrites.copy()
    _type_map = alch.AlchemyExprTranslator._type_map.copy()
    _type_map.update({
        dt.Double: sa.types.REAL,
        dt.Float: sa.types.REAL
    })


rewrites = SQLiteExprTranslator.rewrites
compiles = SQLiteExprTranslator.compiles


class SQLiteDialect(alch.AlchemyDialect):

    translator = SQLiteExprTranslator


dialect = SQLiteDialect



import os

from pkg_resources import parse_version
from pkg_resources.extern.packaging.version import Version

import pytest

import ibis


@pytest.mark.skipif(
    bool(os.environ.get('CIRCLECI', None)),
    reason='Testing import time on CI is flaky due to VM variance',
)
def test_import_time():
    sh = pytest.importorskip('sh')

    lines = [
        'from timeit import timeit',
        "print(timeit('import ibis'))",
    ]

    delta = float(str(sh.python(c='; '.join(lines))))
    assert delta < 2.0


def test_version():
    assert isinstance(parse_version(ibis.__version__), Version)



import ibis
import pytest
import numpy as np
import pandas as pd

from pytest import param
import ibis.tests.util as tu


@pytest.mark.parametrize(('result_fn', 'expected_fn'), [
    param(
        lambda t: t.float_col.lag(),
        lambda t: t.float_col.shift(1),
        id='lag'
    ),
    param(
        lambda t: t.float_col.lead(),
        lambda t: t.float_col.shift(-1),
        id='lead'
    ),
    param(
        lambda t: t.float_col.rank(),
        lambda t: t,
        id='rank',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.float_col.dense_rank(),
        lambda t: t,
        id='dense_rank',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.float_col.percent_rank(),
        lambda t: t,
        id='percent_rank',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.float_col.ntile(buckets=7),
        lambda t: t,
        id='ntile',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.float_col.first(),
        lambda t: t.float_col.head(1),
        id='first',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.float_col.last(),
        lambda t: t.float_col.tail(1),
        id='last',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.float_col.first().over(ibis.window(preceding=10)),
        lambda t: t,
        id='first_preceding',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.float_col.first().over(ibis.window(following=10)),
        lambda t: t,
        id='first_following',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: ibis.row_number(),
        lambda t: pd.Series(np.arange(len(t))),
        id='row_number',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.double_col.cumsum(),
        lambda t: t.double_col.cumsum(),
        id='cumsum'
    ),
    param(
        lambda t: t.double_col.cummean(),
        lambda t: t.double_col.expanding().mean().reset_index(
            drop=True, level=0
        ),
        id='cummean'
    ),
    param(
        lambda t: t.float_col.cummin(),
        lambda t: t.float_col.cummin(),
        id='cummin'
    ),
    param(
        lambda t: t.float_col.cummax(),
        lambda t: t.float_col.cummax(),
        id='cummax'
    ),
    param(
        lambda t: (t.double_col == 0).cumany(),
        lambda t: t.double_col.expanding().agg(
            lambda s: (s == 0).any()
        ).reset_index(drop=True, level=0).astype(bool),
        id='cumany'
    ),
    param(
        lambda t: (t.double_col == 0).cumall(),
        lambda t: t.double_col.expanding().agg(
            lambda s: (s == 0).all()
        ).reset_index(drop=True, level=0).astype(bool),
        id='cumall'
    ),
    param(
        lambda t: t.double_col.sum(),
        lambda gb: gb.double_col.transform('sum'),
        id='sum',
        marks=pytest.mark.xfail,
    ),
    param(
        lambda t: t.double_col.mean(),
        lambda gb: gb.double_col.transform('mean'),
        id='mean'
    ),
    param(
        lambda t: t.float_col.min(),
        lambda gb: gb.float_col.transform('min'),
        id='min'
    ),
    param(
        lambda t: t.float_col.max(),
        lambda gb: gb.float_col.transform('max'),
        id='max'
    ),
])
@tu.skipif_unsupported
def test_window(backend, analytic_alltypes, df, con, result_fn, expected_fn):
    if not backend.supports_window_operations:
        pytest.skip(
            'Backend {} does not support window operations'.format(backend)
        )

    expr = analytic_alltypes.mutate(value=result_fn)

    column = expected_fn(df.sort_values('id').groupby('string_col'))
    expected = (df.assign(value=column)
                  .set_index('id')
                  .sort_index())

    result = expr.execute().set_index('id').sort_index()

    left, right = result.value, expected.value
    backend.assert_series_equal(left, right)



import tempfile
import pytest
import numpy as np
import pandas as pd
import ibis
import ibis.expr.datatypes as dt
from ibis.compat import PY2

pa = pytest.importorskip('pyarrow')  # noqa: E402
import pyarrow.parquet as pq


@pytest.mark.parametrize(('datatype', 'expected'), [
    (pa.int8(), dt.int8),
    (pa.int16(), dt.int16),
    (pa.int32(), dt.int32),
    (pa.int64(), dt.int64),
    (pa.uint8(), dt.uint8),
    (pa.uint16(), dt.uint16),
    (pa.uint32(), dt.uint32),
    (pa.uint64(), dt.uint64),
    (pa.float16(), dt.float16),
    (pa.float32(), dt.float32),
    (pa.float64(), dt.float64),
    (pa.string(), dt.string),
    (pa.timestamp('ns'), dt.timestamp),
], ids=lambda x: str(x))
def test_convert_arrow(datatype, expected):
    assert dt.dtype(datatype) == expected


@pytest.fixture
def parquet_schema():
    np.random.seed(0)
    size = 100
    df = pd.DataFrame({
        'uint8': np.arange(size, dtype=np.uint8),
        'uint16': np.arange(size, dtype=np.uint16),
        'uint32': np.arange(size, dtype=np.uint32),
        'uint64': np.arange(size, dtype=np.uint64),
        'int8': np.arange(size, dtype=np.int16),
        'int16': np.arange(size, dtype=np.int16),
        'int32': np.arange(size, dtype=np.int32),
        'int64': np.arange(size, dtype=np.int64),
        'float32': np.arange(size, dtype=np.float32),
        'float64': np.arange(size, dtype=np.float64),
        'bool': np.random.randn(size) > 0,
        # TODO(wesm): Test other timestamp resolutions now that arrow supports
        # them
        'datetime': np.arange("2016-01-01T00:00:00.001", size,
                              dtype='datetime64[ms]'),
        'str': [str(x) for x in range(size)],
        'str_with_nulls': [None] + [str(x) for x in range(size - 2)] + [None],
        'empty_str': [''] * size,
        'bytes': [b'foo'] * size
        }, columns=['uint8', 'uint16', 'uint32', 'uint64',
                    'int8', 'int16', 'int32',
                    'int64', 'float32', 'float64', 'bool',
                    'datetime', 'str', 'str_with_nulls', 'empty_str',
                    'bytes'])

    with tempfile.TemporaryFile() as path:
        table = pa.Table.from_pandas(df)
        pq.write_table(table, path)
        parquet_file = pq.ParquetFile(path)
        return parquet_file.schema


def test_convert_parquet(parquet_schema):
    # TODO(jreback)
    # not entirely sure this is correct
    # should these be strings in py2?
    if PY2:
        strings = [dt.binary, dt.binary, dt.binary]
    else:
        strings = [dt.string, dt.string, dt.string]

    # uint32, int8, int16 stored as upcasted types
    types = [dt.uint8, dt.uint16, dt.int64, dt.uint64,
             dt.int16, dt.int16, dt.int32,
             dt.int64, dt.float32, dt.float64, dt.boolean,
             dt.timestamp] + strings + [dt.binary, dt.int64]
    names = ['uint8', 'uint16', 'uint32', 'uint64',
             'int8', 'int16', 'int32',
             'int64', 'float32', 'float64', 'bool',
             'datetime', 'str', 'str_with_nulls', 'empty_str',
             'bytes', '__index_level_0__']
    expected = ibis.schema(zip(names, types))

    result = ibis.infer_schema(parquet_schema)
    assert result == expected



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ibis.client import SQLClient
from ibis.expr.schema import Schema


class MockConnection(SQLClient):

    @property
    def dialect(self):
        from ibis.impala.compiler import ImpalaDialect
        return ImpalaDialect

    _tables = {
        'alltypes': [
            ('a', 'int8'),
            ('b', 'int16'),
            ('c', 'int32'),
            ('d', 'int64'),
            ('e', 'float'),
            ('f', 'double'),
            ('g', 'string'),
            ('h', 'boolean'),
            ('i', 'timestamp'),
            ('j', 'date'),
            ('k', 'time')
        ],
        'star1': [
            ('c', 'int32'),
            ('f', 'double'),
            ('foo_id', 'string'),
            ('bar_id', 'string'),
        ],
        'star2': [
            ('foo_id', 'string'),
            ('value1', 'double'),
            ('value3', 'double')
        ],
        'star3': [
            ('bar_id', 'string'),
            ('value2', 'double')
        ],
        'test1': [
            ('c', 'int32'),
            ('f', 'double'),
            ('g', 'string')
        ],
        'test2': [
            ('key', 'string'),
            ('value', 'double')
        ],
        'tpch_region': [
            ('r_regionkey', 'int16'),
            ('r_name', 'string'),
            ('r_comment', 'string')
        ],
        'tpch_nation': [
            ('n_nationkey', 'int16'),
            ('n_name', 'string'),
            ('n_regionkey', 'int16'),
            ('n_comment', 'string')
        ],
        'tpch_lineitem': [
            ('l_orderkey', 'int64'),
            ('l_partkey', 'int64'),
            ('l_suppkey', 'int64'),
            ('l_linenumber', 'int32'),
            ('l_quantity', 'decimal(12,2)'),
            ('l_extendedprice', 'decimal(12,2)'),
            ('l_discount', 'decimal(12,2)'),
            ('l_tax', 'decimal(12,2)'),
            ('l_returnflag', 'string'),
            ('l_linestatus', 'string'),
            ('l_shipdate', 'string'),
            ('l_commitdate', 'string'),
            ('l_receiptdate', 'string'),
            ('l_shipinstruct', 'string'),
            ('l_shipmode', 'string'),
            ('l_comment', 'string')
        ],
        'tpch_customer': [
            ('c_custkey', 'int64'),
            ('c_name', 'string'),
            ('c_address', 'string'),
            ('c_nationkey', 'int16'),
            ('c_phone', 'string'),
            ('c_acctbal', 'decimal'),
            ('c_mktsegment', 'string'),
            ('c_comment', 'string')
        ],
        'tpch_orders': [
            ('o_orderkey', 'int64'),
            ('o_custkey', 'int64'),
            ('o_orderstatus', 'string'),
            ('o_totalprice', 'decimal(12,2)'),
            ('o_orderdate', 'string'),
            ('o_orderpriority', 'string'),
            ('o_clerk', 'string'),
            ('o_shippriority', 'int32'),
            ('o_comment', 'string')
        ],
        'functional_alltypes': [
            ('id', 'int32'),
            ('bool_col', 'boolean'),
            ('tinyint_col', 'int8'),
            ('smallint_col', 'int16'),
            ('int_col', 'int32'),
            ('bigint_col', 'int64'),
            ('float_col', 'float'),
            ('double_col', 'double'),
            ('date_string_col', 'string'),
            ('string_col', 'string'),
            ('timestamp_col', 'timestamp'),
            ('year', 'int32'),
            ('month', 'int32')
        ],
        'airlines': [
            ('year', 'int32'),
            ('month', 'int32'),
            ('day', 'int32'),
            ('dayofweek', 'int32'),
            ('dep_time', 'int32'),
            ('crs_dep_time', 'int32'),
            ('arr_time', 'int32'),
            ('crs_arr_time', 'int32'),
            ('carrier', 'string'),
            ('flight_num', 'int32'),
            ('tail_num', 'int32'),
            ('actual_elapsed_time', 'int32'),
            ('crs_elapsed_time', 'int32'),
            ('airtime', 'int32'),
            ('arrdelay', 'int32'),
            ('depdelay', 'int32'),
            ('origin', 'string'),
            ('dest', 'string'),
            ('distance', 'int32'),
            ('taxi_in', 'int32'),
            ('taxi_out', 'int32'),
            ('cancelled', 'int32'),
            ('cancellation_code', 'string'),
            ('diverted', 'int32'),
            ('carrier_delay', 'int32'),
            ('weather_delay', 'int32'),
            ('nas_delay', 'int32'),
            ('security_delay', 'int32'),
            ('late_aircraft_delay', 'int32')
        ],
        'tpcds_customer': [
            ('c_customer_sk', 'int64'),
            ('c_customer_id', 'string'),
            ('c_current_cdemo_sk', 'int32'),
            ('c_current_hdemo_sk', 'int32'),
            ('c_current_addr_sk', 'int32'),
            ('c_first_shipto_date_sk', 'int32'),
            ('c_first_sales_date_sk', 'int32'),
            ('c_salutation', 'string'),
            ('c_first_name', 'string'),
            ('c_last_name', 'string'),
            ('c_preferred_cust_flag', 'string'),
            ('c_birth_day', 'int32'),
            ('c_birth_month', 'int32'),
            ('c_birth_year', 'int32'),
            ('c_birth_country', 'string'),
            ('c_login', 'string'),
            ('c_email_address', 'string'),
            ('c_last_review_date', 'string')],
        'tpcds_customer_address': [
            ('ca_address_sk', 'bigint'),
            ('ca_address_id', 'string'),
            ('ca_street_number', 'string'),
            ('ca_street_name', 'string'),
            ('ca_street_type', 'string'),
            ('ca_suite_number', 'string'),
            ('ca_city', 'string'),
            ('ca_county', 'string'),
            ('ca_state', 'string'),
            ('ca_zip', 'string'),
            ('ca_country', 'string'),
            ('ca_gmt_offset', 'decimal(5,2)'),
            ('ca_location_type', 'string')],
        'tpcds_customer_demographics': [
            ('cd_demo_sk', 'bigint'),
            ('cd_gender', 'string'),
            ('cd_marital_status', 'string'),
            ('cd_education_status', 'string'),
            ('cd_purchase_estimate', 'int'),
            ('cd_credit_rating', 'string'),
            ('cd_dep_count', 'int'),
            ('cd_dep_employed_count', 'int'),
            ('cd_dep_college_count', 'int')],
        'tpcds_date_dim': [
            ('d_date_sk', 'bigint'),
            ('d_date_id', 'string'),
            ('d_date', 'string'),
            ('d_month_seq', 'int'),
            ('d_week_seq', 'int'),
            ('d_quarter_seq', 'int'),
            ('d_year', 'int'),
            ('d_dow', 'int'),
            ('d_moy', 'int'),
            ('d_dom', 'int'),
            ('d_qoy', 'int'),
            ('d_fy_year', 'int'),
            ('d_fy_quarter_seq', 'int'),
            ('d_fy_week_seq', 'int'),
            ('d_day_name', 'string'),
            ('d_quarter_name', 'string'),
            ('d_holiday', 'string'),
            ('d_weekend', 'string'),
            ('d_following_holiday', 'string'),
            ('d_first_dom', 'int'),
            ('d_last_dom', 'int'),
            ('d_same_day_ly', 'int'),
            ('d_same_day_lq', 'int'),
            ('d_current_day', 'string'),
            ('d_current_week', 'string'),
            ('d_current_month', 'string'),
            ('d_current_quarter', 'string'),
            ('d_current_year', 'string')],
        'tpcds_household_demographics': [
            ('hd_demo_sk', 'bigint'),
            ('hd_income_band_sk', 'int'),
            ('hd_buy_potential', 'string'),
            ('hd_dep_count', 'int'),
            ('hd_vehicle_count', 'int')],
        'tpcds_item': [
            ('i_item_sk', 'bigint'),
            ('i_item_id', 'string'),
            ('i_rec_start_date', 'string'),
            ('i_rec_end_date', 'string'),
            ('i_item_desc', 'string'),
            ('i_current_price', 'decimal(7,2)'),
            ('i_wholesale_cost', 'decimal(7,2)'),
            ('i_brand_id', 'int'),
            ('i_brand', 'string'),
            ('i_class_id', 'int'),
            ('i_class', 'string'),
            ('i_category_id', 'int'),
            ('i_category', 'string'),
            ('i_manufact_id', 'int'),
            ('i_manufact', 'string'),
            ('i_size', 'string'),
            ('i_formulation', 'string'),
            ('i_color', 'string'),
            ('i_units', 'string'),
            ('i_container', 'string'),
            ('i_manager_id', 'int'),
            ('i_product_name', 'string')],
        'tpcds_promotion': [
            ('p_promo_sk', 'bigint'),
            ('p_promo_id', 'string'),
            ('p_start_date_sk', 'int'),
            ('p_end_date_sk', 'int'),
            ('p_item_sk', 'int'),
            ('p_cost', 'decimal(15,2)'),
            ('p_response_target', 'int'),
            ('p_promo_name', 'string'),
            ('p_channel_dmail', 'string'),
            ('p_channel_email', 'string'),
            ('p_channel_catalog', 'string'),
            ('p_channel_tv', 'string'),
            ('p_channel_radio', 'string'),
            ('p_channel_press', 'string'),
            ('p_channel_event', 'string'),
            ('p_channel_demo', 'string'),
            ('p_channel_details', 'string'),
            ('p_purpose', 'string'),
            ('p_discount_active', 'string')],
        'tpcds_store': [
            ('s_store_sk', 'bigint'),
            ('s_store_id', 'string'),
            ('s_rec_start_date', 'string'),
            ('s_rec_end_date', 'string'),
            ('s_closed_date_sk', 'int'),
            ('s_store_name', 'string'),
            ('s_number_employees', 'int'),
            ('s_floor_space', 'int'),
            ('s_hours', 'string'),
            ('s_manager', 'string'),
            ('s_market_id', 'int'),
            ('s_geography_class', 'string'),
            ('s_market_desc', 'string'),
            ('s_market_manager', 'string'),
            ('s_division_id', 'int'),
            ('s_division_name', 'string'),
            ('s_company_id', 'int'),
            ('s_company_name', 'string'),
            ('s_street_number', 'string'),
            ('s_street_name', 'string'),
            ('s_street_type', 'string'),
            ('s_suite_number', 'string'),
            ('s_city', 'string'),
            ('s_county', 'string'),
            ('s_state', 'string'),
            ('s_zip', 'string'),
            ('s_country', 'string'),
            ('s_gmt_offset', 'decimal(5,2)'),
            ('s_tax_precentage', 'decimal(5,2)')],
        'tpcds_store_sales': [
            ('ss_sold_time_sk', 'bigint'),
            ('ss_item_sk', 'bigint'),
            ('ss_customer_sk', 'bigint'),
            ('ss_cdemo_sk', 'bigint'),
            ('ss_hdemo_sk', 'bigint'),
            ('ss_addr_sk', 'bigint'),
            ('ss_store_sk', 'bigint'),
            ('ss_promo_sk', 'bigint'),
            ('ss_ticket_number', 'int'),
            ('ss_quantity', 'int'),
            ('ss_wholesale_cost', 'decimal(7,2)'),
            ('ss_list_price', 'decimal(7,2)'),
            ('ss_sales_price', 'decimal(7,2)'),
            ('ss_ext_discount_amt', 'decimal(7,2)'),
            ('ss_ext_sales_price', 'decimal(7,2)'),
            ('ss_ext_wholesale_cost', 'decimal(7,2)'),
            ('ss_ext_list_price', 'decimal(7,2)'),
            ('ss_ext_tax', 'decimal(7,2)'),
            ('ss_coupon_amt', 'decimal(7,2)'),
            ('ss_net_paid', 'decimal(7,2)'),
            ('ss_net_paid_inc_tax', 'decimal(7,2)'),
            ('ss_net_profit', 'decimal(7,2)'),
            ('ss_sold_date_sk', 'bigint')],
        'tpcds_time_dim': [
            ('t_time_sk', 'bigint'),
            ('t_time_id', 'string'),
            ('t_time', 'int'),
            ('t_hour', 'int'),
            ('t_minute', 'int'),
            ('t_second', 'int'),
            ('t_am_pm', 'string'),
            ('t_shift', 'string'),
            ('t_sub_shift', 'string'),
            ('t_meal_time', 'string')]
    }

    def __init__(self):
        self.executed_queries = []

    def _get_table_schema(self, name):
        name = name.replace('`', '')
        return Schema.from_tuples(self._tables[name])

    def _build_ast(self, expr, context):
        from ibis.impala.compiler import build_ast
        return build_ast(expr, context)

    def execute(self, expr, limit=None, _async=False, params=None):
        if _async:
            raise NotImplementedError
        ast = self._build_ast_ensure_limit(expr, limit, params=params)
        for query in ast.queries:
            self.executed_queries.append(query.compile())
        return None

    def compile(self, expr, limit=None, params=None):
        ast = self._build_ast_ensure_limit(expr, limit, params=params)
        queries = [q.compile() for q in ast.queries]
        return queries[0] if len(queries) == 1 else queries



import os

import pytest

import ibis


PROJECT_ID = os.environ.get('GOOGLE_BIGQUERY_PROJECT_ID')
DATASET_ID = 'testing'


@pytest.fixture(scope='session')
def client():
    ga = pytest.importorskip('google.auth')

    try:
        return ibis.bigquery.connect(PROJECT_ID, DATASET_ID)
    except ga.exceptions.DefaultCredentialsError:
        pytest.skip("no credentials found, skipping")


@pytest.fixture(scope='session')
def alltypes(client):
    return client.table('functional_alltypes')


@pytest.fixture(scope='session')
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope='session')
def parted_alltypes(client):
    return client.table('functional_alltypes_parted')


@pytest.fixture(scope='session')
def parted_df(parted_alltypes):
    return parted_alltypes.execute()


@pytest.fixture(scope='session')
def struct_table(client):
    return client.table('struct_table')



import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

import ibis.expr.schema as sch
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops

from ibis.compat import parse_version
from ibis.file.client import FileClient
from ibis.pandas.api import PandasDialect
from ibis.pandas.core import pre_execute, execute


dialect = PandasDialect


# TODO(jreback) complex types are not implemented
_arrow_dtypes = {
    'int8': dt.Int8,
    'int16': dt.Int16,
    'int32': dt.Int32,
    'int64': dt.Int64,
    'uint8': dt.UInt8,
    'uint16': dt.UInt16,
    'uint32': dt.UInt32,
    'uint64': dt.UInt64,
    'halffloat': dt.Float16,
    'float': dt.Float32,
    'double': dt.Float64,
    'string': dt.String,
    'binary': dt.Binary,
    'bool': dt.Boolean,
    'timestamp[ns]': dt.Timestamp,
    'timestamp[us]': dt.Timestamp
}


@dt.dtype.register(pa.DataType)
def pa_dtype(arrow_type, nullable=True):
    return _arrow_dtypes[str(arrow_type)](nullable=nullable)


@sch.infer.register(pq.ParquetSchema)
def infer_parquet_schema(schema):
    pairs = []

    for field in schema.to_arrow_schema():
        ibis_dtype = dt.dtype(field.type, nullable=field.nullable)
        pairs.append((field.name, ibis_dtype))

    return sch.schema(pairs)


def connect(dictionary):
    return ParquetClient(dictionary)


class ParquetTable(ops.DatabaseTable):
    pass


class ParquetClient(FileClient):

    dialect = dialect
    extension = 'parquet'

    def insert(self, path, expr, **kwargs):

        path = self.root / path
        df = execute(expr)
        table = pa.Table.from_pandas(df)
        pq.write_table(table, str(path))

    def table(self, name, path):
        if name not in self.list_tables(path):
            raise AttributeError(name)

        if path is None:
            path = self.root

        # get the schema
        f = path / "{}.parquet".format(name)

        parquet_file = pq.ParquetFile(str(f))
        schema = sch.infer(parquet_file.schema)

        table = ParquetTable(name, schema, self).to_expr()
        self.dictionary[name] = f

        return table

    def list_tables(self, path=None):
        return self._list_tables_files(path)

    def list_databases(self, path=None):
        return self._list_databases_dirs(path)

    def compile(self, expr, *args, **kwargs):
        return expr

    @property
    def version(self):
        return parse_version(pa.__version__)


@pre_execute.register(ParquetTable, ParquetClient)
def parquet_pre_execute_client(op, client, scope, **kwargs):
    # cache
    if isinstance(scope.get(op), pd.DataFrame):
        return {}

    path = client.dictionary[op.name]
    table = pq.read_table(str(path))
    df = table.to_pandas()
    return {op: df}



import pytest

import pandas as pd
import pandas.util.testing as tm

import ibis

pytest.importorskip('multipledispatch')

from ibis.pandas.client import PandasTable  # noqa: E402

pytestmark = pytest.mark.pandas


@pytest.fixture
def client():
    return ibis.pandas.connect({
        'df': pd.DataFrame({'a': [1, 2, 3], 'b': list('abc')}),
        'df_unknown': pd.DataFrame({
            'array_of_strings': [['a', 'b'], [], ['c']],
        }),
    })


@pytest.fixture
def table(client):
    return client.table('df')


def test_client_table(table):
    assert isinstance(table.op(), ibis.expr.operations.DatabaseTable)
    assert isinstance(table.op(), PandasTable)


def test_client_table_repr(table):
    assert 'PandasTable' in repr(table)


def test_literal(client):
    assert client.execute(ibis.literal(1)) == 1


def test_read_with_undiscoverable_type(client):
    with pytest.raises(TypeError):
        client.table('df_unknown')


def test_drop(table):
    table = table.mutate(c=table.a)
    expr = table.drop(['a'])
    result = expr.execute()
    expected = table[['b', 'c']].execute()
    tm.assert_frame_equal(result, expected)



import os
import pytest

from ibis.compat import Path
from ibis.tests.backends import (Csv, Parquet, Pandas,
                                 SQLite, Postgres, MySQL,
                                 Clickhouse, Impala, BigQuery)


pytestmark = pytest.mark.backend


@pytest.fixture(scope='session')
def data_directory():
    root = Path(__file__).absolute().parents[3]

    default = root / 'ci' / 'ibis-testing-data'
    datadir = os.environ.get('IBIS_TEST_DATA_DIRECTORY', default)
    datadir = Path(datadir)

    if not datadir.exists():
        pytest.skip('test data directory not found')

    return datadir


@pytest.fixture(params=[
    pytest.param(Csv, marks=pytest.mark.csv),
    pytest.param(Parquet, marks=pytest.mark.parquet),
    pytest.param(Pandas, marks=pytest.mark.pandas),
    pytest.param(SQLite, marks=pytest.mark.sqlite),
    pytest.param(Postgres, marks=pytest.mark.postgres),
    pytest.param(MySQL, marks=pytest.mark.mysql),
    pytest.param(Clickhouse, marks=pytest.mark.clickhouse),
    pytest.param(BigQuery, marks=pytest.mark.bigquery),
    pytest.param(Impala, marks=pytest.mark.impala)
], scope='session')
def backend(request, data_directory):
    return request.param(data_directory)


@pytest.fixture(scope='session')
def con(backend):
    return backend.connection


@pytest.fixture(scope='session')
def alltypes(backend):
    return backend.functional_alltypes()


@pytest.fixture(scope='session')
def batting(backend):
    return backend.batting()


@pytest.fixture(scope='session')
def awards_players(backend):
    return backend.awards_players()


@pytest.fixture
def analytic_alltypes(alltypes):
    return alltypes.groupby('string_col').order_by('id')


@pytest.fixture(scope='session')
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope='session')
def batting_df(batting):
    return batting.execute(limit=None)


@pytest.fixture(scope='session')
def awards_players_df(awards_players):
    return awards_players.execute(limit=None)



import getpass
import pymysql  # NOQA fail early if the driver is missing
import warnings
import contextlib
import sqlalchemy as sa
import sqlalchemy.dialects.mysql as mysql

from ibis.sql.mysql.compiler import MySQLDialect
import ibis.sql.alchemy as alch
import ibis.expr.datatypes as dt


# TODO(kszucs): unsigned integers

@dt.dtype.register((mysql.DOUBLE, mysql.REAL))
def mysql_double(satype, nullable=True):
    return dt.Double(nullable=nullable)


@dt.dtype.register(mysql.FLOAT)
def mysql_float(satype, nullable=True):
    return dt.Float(nullable=nullable)


@dt.dtype.register(mysql.TINYINT)
def mysql_tinyint(satype, nullable=True):
    return dt.Int8(nullable=nullable)


@dt.dtype.register(mysql.BLOB)
def mysql_blob(satype, nullable=True):
    return dt.Binary(nullable=nullable)


class MySQLTable(alch.AlchemyTable):
    pass


class MySQLSchema(alch.AlchemyDatabaseSchema):
    pass


class MySQLDatabase(alch.AlchemyDatabase):
    schema_class = MySQLSchema


class MySQLClient(alch.AlchemyClient):

    """The Ibis MySQL client class

    Attributes
    ----------
    con : sqlalchemy.engine.Engine
    """

    dialect = MySQLDialect
    database_class = MySQLDatabase

    def __init__(self, host='localhost', user=None, password=None, port=3306,
                 database='mysql', url=None, driver='pymysql'):
        if url is None:
            if driver != 'pymysql':
                raise NotImplementedError(
                    'pymysql is currently the only supported driver'
                )
            user = user or getpass.getuser()
            url = sa.engine.url.URL('mysql+pymysql', host=host, port=port,
                                    username=user, password=password,
                                    database=database)
        else:
            url = sa.engine.url.make_url(url)

        super(MySQLClient, self).__init__(sa.create_engine(url))
        self.database_name = url.database

    @contextlib.contextmanager
    def begin(self):
        with super(MySQLClient, self).begin() as bind:
            previous_timezone = (bind.execute('SELECT @@session.time_zone')
                                     .scalar())
            try:
                bind.execute("SET @@session.time_zone = 'UTC'")
            except Exception as e:
                warnings.warn("Couldn't set mysql timezone: {}".format(str(e)))

            try:
                yield bind
            finally:
                query = "SET @@session.time_zone = '{}'"
                bind.execute(query.format(previous_timezone))

    def database(self, name=None):
        """Connect to a database called `name`.

        Parameters
        ----------
        name : str, optional
            The name of the database to connect to. If ``None``, return
            the database named ``self.current_database``.

        Returns
        -------
        db : MySQLDatabase
            An :class:`ibis.sql.mysql.client.MySQLDatabase` instance.

        Notes
        -----
        This creates a new connection if `name` is both not ``None`` and not
        equal to the current database.
        """
        if name == self.current_database or (
            name is None and name != self.current_database
        ):
            return self.database_class(self.current_database, self)
        else:
            url = self.con.url
            client_class = type(self)
            new_client = client_class(
                host=url.host,
                user=url.username,
                port=url.port,
                password=url.password,
                database=name,
            )
            return self.database_class(name, new_client)

    def schema(self, name):
        """Get a schema object from the current database for the schema named `name`.

        Parameters
        ----------
        name : str

        Returns
        -------
        schema : MySQLSchema
            An :class:`ibis.sql.mysql.client.MySQLSchema` instance.

        """
        return self.database().schema(name)

    @property
    def current_database(self):
        """The name of the current database this client is connected to."""
        return self.database_name

    def list_databases(self):
        return [row.Database for row in self.con.execute('SHOW DATABASES')]

    def list_schemas(self):
        """List all the schemas in the current database."""
        return self.inspector.get_schema_names()

    def set_database(self, name):
        raise NotImplementedError(
            'Cannot set database with MySQL client. To use a different'
            ' database, use client.database({!r})'.format(name)
        )

    @property
    def client(self):
        return self

    def table(self, name, database=None, schema=None):
        """Create a table expression that references a particular a table
        called `name` in a MySQL database called `database`.

        Parameters
        ----------
        name : str
            The name of the table to retrieve.
        database : str, optional
            The database in which the table referred to by `name` resides. If
            ``None`` then the ``current_database`` is used.
        schema : str, optional
            The schema in which the table resides.  If ``None`` then the
            `public` schema is assumed.

        Returns
        -------
        table : TableExpr
            A table expression.
        """
        if database is not None and database != self.current_database:
            return (
                self.database(name=database)
                    .table(name=name, schema=schema)
            )
        else:
            alch_table = self._get_sqla_table(name, schema=schema)
            node = MySQLTable(alch_table, self, self._schemas.get(name))
            return self._table_expr_klass(node)

    def list_tables(self, like=None, database=None, schema=None):
        if database is not None and database != self.current_database:
            return (
                self.database(name=database)
                    .list_tables(like=like, schema=schema)
            )
        else:
            parent = super(MySQLClient, self)
            return parent.list_tables(like=like, schema=schema)



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import toolz

import ibis.expr.types as ir
import ibis.expr.lineage as lin
import ibis.expr.operations as ops

from ibis.expr.schema import HasSchema
from ibis.expr.window import window

from ibis.common import RelationError, ExpressionError, IbisTypeError

# ---------------------------------------------------------------------
# Some expression metaprogramming / graph transformations to support
# compilation later


def sub_for(expr, substitutions):
    mapping = {repr(k.op()): v for k, v in substitutions}
    substitutor = Substitutor()
    return substitutor.substitute(expr, mapping)


def _expr_key(expr):
    try:
        name = expr.get_name()
    except (AttributeError, ExpressionError):
        name = None

    try:
        op = expr.op()
    except AttributeError:
        return expr, name
    else:
        return repr(op), name


class Substitutor(object):

    def __init__(self):
        cache = toolz.memoize(key=lambda args, kwargs: _expr_key(args[0]))
        self.substitute = cache(self._substitute)

    def _substitute(self, expr, mapping):
        """Substitute expressions with other expressions.

        Parameters
        ----------
        expr : ibis.expr.types.Expr
        mapping : Dict, OrderedDict

        Returns
        -------
        new_expr : ibis.expr.types.Expr
        """
        node = expr.op()
        key = repr(node)
        if key in mapping:
            return mapping[key]
        if node.blocks():
            return expr

        new_args = list(node.args)
        unchanged = True
        for i, arg in enumerate(new_args):
            if isinstance(arg, ir.Expr):
                new_arg = self.substitute(arg, mapping)
                unchanged = unchanged and new_arg is arg
                new_args[i] = new_arg
        if unchanged:
            return expr
        try:
            new_node = type(node)(*new_args)
        except IbisTypeError:
            return expr

        try:
            name = expr.get_name()
        except ExpressionError:
            name = None
        return expr._factory(new_node, name=name)


class ScalarAggregate(object):

    def __init__(self, expr, memo=None, default_name='tmp'):
        self.expr = expr
        self.memo = memo or {}
        self.tables = []
        self.default_name = default_name

    def get_result(self):
        expr = self.expr
        subbed_expr = self._visit(expr)

        try:
            name = subbed_expr.get_name()
            named_expr = subbed_expr
        except ExpressionError:
            name = self.default_name
            named_expr = subbed_expr.name(self.default_name)

        table = self.tables[0]
        for other in self.tables[1:]:
            table = table.cross_join(other)

        return table.projection([named_expr]), name

    def _visit(self, expr):
        if is_scalar_reduce(expr) and not has_multiple_bases(expr):
            # An aggregation unit
            key = self._key(expr)
            if key not in self.memo:
                agg_expr, name = reduction_to_aggregation(expr)
                self.memo[key] = agg_expr, name
                self.tables.append(agg_expr)
            else:
                agg_expr, name = self.memo[key]
            return agg_expr[name]

        elif not isinstance(expr, ir.Expr):
            return expr

        node = expr.op()
        subbed_args = []
        for arg in node.args:
            if isinstance(arg, (tuple, list)):
                subbed_arg = [self._visit(x) for x in arg]
            else:
                subbed_arg = self._visit(arg)
            subbed_args.append(subbed_arg)

        subbed_node = type(node)(*subbed_args)
        if isinstance(expr, ir.ValueExpr):
            result = expr._factory(subbed_node, name=expr._name)
        else:
            result = expr._factory(subbed_node)

        return result

    def _key(self, expr):
        return repr(expr.op())


def has_multiple_bases(expr):
    return toolz.count(find_immediate_parent_tables(expr)) > 1


def reduction_to_aggregation(expr, default_name='tmp'):
    tables = list(find_immediate_parent_tables(expr))

    try:
        name = expr.get_name()
        named_expr = expr
    except ExpressionError:
        name = default_name
        named_expr = expr.name(default_name)

    if len(tables) == 1:
        table, = tables
        return table.aggregate([named_expr]), name
    else:
        return ScalarAggregate(expr, None, default_name).get_result()


def find_immediate_parent_tables(expr):
    """Find every first occurrence of a :class:`ibis.expr.types.TableExpr`
    object in `expr`.

    Parameters
    ----------
    expr : ir.Expr

    Yields
    ------
    e : ir.Expr

    Notes
    -----
    This function does not traverse into TableExpr objects. This means that the
    underlying PhysicalTable of a Selection will not be yielded, for example.

    Examples
    --------
    >>> import ibis, toolz
    >>> t = ibis.table([('a', 'int64')], name='t')
    >>> expr = t.mutate(foo=t.a + 1)
    >>> result = list(find_immediate_parent_tables(expr))
    >>> len(result)
    1
    >>> result[0]  # doctest: +NORMALIZE_WHITESPACE
    ref_0
    UnboundTable[table]
      name: t
      schema:
        a : int64
    Selection[table]
      table:
        Table: ref_0
      selections:
        Table: ref_0
        foo = Add[int64*]
          left:
            a = Column[int64*] 'a' from table
              ref_0
          right:
            Literal[int8]
              1
    """
    def finder(expr):
        if isinstance(expr, ir.TableExpr):
            return lin.halt, expr
        else:
            return lin.proceed, None

    return lin.traverse(finder, expr)


def is_scalar_reduce(x):
    return isinstance(x, ir.ScalarExpr) and ops.is_reduction(x)


def substitute_parents(expr, lift_memo=None, past_projection=True):
    rewriter = ExprSimplifier(expr, lift_memo=lift_memo,
                              block_projection=not past_projection)
    return rewriter.get_result()


class ExprSimplifier(object):

    """
    Rewrite the input expression by replacing any table expressions part of a
    "commutative table operation unit" (for lack of scientific term, a set of
    operations that can be written down in any order and still yield the same
    semantic result)
    """

    def __init__(self, expr, lift_memo=None, block_projection=False):
        self.expr = expr
        self.lift_memo = lift_memo or {}

        self.block_projection = block_projection

    def get_result(self):
        expr = self.expr
        node = expr.op()
        if isinstance(node, ir.Literal):
            return expr

        # For table column references, in the event that we're on top of a
        # projection, we need to check whether the ref comes from the base
        # table schema or is a derived field. If we've projected out of
        # something other than a physical table, then lifting should not occur
        if isinstance(node, ops.TableColumn):
            result = self._lift_TableColumn(expr, block=self.block_projection)
            if result is not expr:
                return result
        # Temporary hacks around issues addressed in #109
        elif isinstance(node, ops.Selection):
            return self._lift_Selection(expr, block=self.block_projection)
        elif isinstance(node, ops.Aggregation):
            return self._lift_Aggregation(expr, block=self.block_projection)

        unchanged = True

        lifted_args = []
        for arg in node.args:
            lifted_arg, unch_arg = self._lift_arg(
                arg, block=self.block_projection)
            lifted_args.append(lifted_arg)

            unchanged = unchanged and unch_arg

        # Do not modify unnecessarily
        if unchanged:
            return expr

        lifted_node = type(node)(*lifted_args)
        if isinstance(expr, ir.ValueExpr):
            result = expr._factory(lifted_node, name=expr._name)
        else:
            result = expr._factory(lifted_node)

        return result

    def _lift_arg(self, arg, block=None):
        unchanged = [True]

        def _lift(x):
            if isinstance(x, ir.Expr):
                lifted_arg = self.lift(x, block=block)
                if lifted_arg is not x:
                    unchanged[0] = False
            else:
                # a string or some other thing
                lifted_arg = x
            return lifted_arg

        if arg is None:
            return arg, True

        if isinstance(arg, (tuple, list)):
            result = [_lift(x) for x in arg]
        else:
            result = _lift(arg)

        return result, unchanged[0]

    def lift(self, expr, block=None):
        # This use of id() is OK since only for memoization
        key = id(expr.op()), block

        if key in self.lift_memo:
            return self.lift_memo[key]

        op = expr.op()

        if isinstance(op, ops.ValueOp):
            return self._sub(expr, block=block)
        elif isinstance(op, ops.Selection):
            result = self._lift_Selection(expr, block=block)
        elif isinstance(op, ops.Join):
            result = self._lift_Join(expr, block=block)
        elif isinstance(op, (ops.TableNode, HasSchema)):
            return expr
        else:
            return self._sub(expr, block=block)

        # If we get here, time to record the modified expression in our memo to
        # avoid excessive graph-walking
        self.lift_memo[key] = result
        return result

    def _lift_TableColumn(self, expr, block=None):
        node = expr.op()

        tnode = node.table.op()
        root = _base_table(tnode)

        result = expr
        if isinstance(root, ops.Selection):
            can_lift = False

            for val in root.selections:
                if (isinstance(val.op(), ops.PhysicalTable) and
                        node.name in val.schema()):

                    can_lift = True
                    lifted_root = self.lift(val)
                elif (isinstance(val.op(), ops.TableColumn) and
                      val.op().name == val.get_name() and
                      node.name == val.get_name()):
                    can_lift = True
                    lifted_root = self.lift(val.op().table)

                # XXX
                # can_lift = False

            # HACK: If we've projected a join, do not lift the children
            # TODO: what about limits and other things?
            # if isinstance(root.table.op(), Join):
            #     can_lift = False

            if can_lift and not block:
                lifted_node = ops.TableColumn(node.name, lifted_root)
                result = expr._factory(lifted_node, name=expr._name)

        return result

    def _lift_Aggregation(self, expr, block=None):
        if block is None:
            block = self.block_projection

        op = expr.op()

        # as exposed in #544, do not lift the table inside (which may be
        # filtered or otherwise altered in some way) if blocking

        if block:
            lifted_table = op.table
        else:
            lifted_table = self.lift(op.table, block=True)

        unch = lifted_table is op.table

        lifted_aggs, unch1 = self._lift_arg(op.metrics, block=True)
        lifted_by, unch2 = self._lift_arg(op.by, block=True)
        lifted_having, unch3 = self._lift_arg(op.having, block=True)

        unchanged = unch and unch1 and unch2 and unch3

        if not unchanged:
            lifted_op = ops.Aggregation(lifted_table, lifted_aggs,
                                        by=lifted_by, having=lifted_having)
            result = ir.TableExpr(lifted_op)
        else:
            result = expr

        return result

    def _lift_Selection(self, expr, block=None):
        if block is None:
            block = self.block_projection

        op = expr.op()

        if block and op.blocks():
            # GH #549: dig no further
            return expr
        else:
            lifted_table, unch = self._lift_arg(op.table, block=True)

        lifted_selections, unch_sel = self._lift_arg(op.selections, block=True)
        unchanged = unch and unch_sel

        lifted_predicates, unch_sel = self._lift_arg(op.predicates, block=True)
        unchanged = unch and unch_sel

        lifted_sort_keys, unch_sel = self._lift_arg(op.sort_keys, block=True)
        unchanged = unch and unch_sel

        if not unchanged:
            lifted_projection = ops.Selection(lifted_table, lifted_selections,
                                              lifted_predicates,
                                              lifted_sort_keys)
            result = ir.TableExpr(lifted_projection)
        else:
            result = expr

        return result

    def _lift_Join(self, expr, block=None):
        op = expr.op()

        left_lifted = self.lift(op.left, block=block)
        right_lifted = self.lift(op.right, block=block)

        unchanged = (left_lifted is op.left and
                     right_lifted is op.right)

        # Fix predicates
        lifted_preds = []
        for x in op.predicates:
            subbed = self._sub(x, block=True)
            if subbed is not x:
                unchanged = False
            lifted_preds.append(subbed)

        if not unchanged:
            lifted_join = type(op)(left_lifted, right_lifted, lifted_preds)
            result = ir.TableExpr(lifted_join)
        else:
            result = expr

        return result

    def _sub(self, expr, block=None):
        # catchall recursive rewriter
        if block is None:
            block = self.block_projection

        helper = ExprSimplifier(expr, lift_memo=self.lift_memo,
                                block_projection=block)
        return helper.get_result()


def _base_table(table_node):
    # Find the aggregate or projection root. Not proud of this
    if table_node.blocks():
        return table_node
    else:
        return _base_table(table_node.table.op())


def has_reduction(expr):
    """Does `expr` contain a reduction?

    Parameters
    ----------
    expr : ibis.expr.types.Expr
        An ibis expression

    Returns
    -------
    truth_value : bool
        Whether or not there's at least one reduction in `expr`

    Notes
    -----
    The ``isinstance(op, ops.TableNode)`` check in this function implies
    that we only examine every non-table expression that precedes the first
    table expression.
    """
    def fn(expr):
        op = expr.op()
        if isinstance(op, ops.TableNode):  # don't go below any table nodes
            return lin.halt, None
        if isinstance(op, ops.Reduction):
            return lin.halt, True
        return lin.proceed, None
    reduction_status = lin.traverse(fn, expr)
    return any(reduction_status)


def apply_filter(expr, predicates):
    # This will attempt predicate pushdown in the cases where we can do it
    # easily and safely, to make both cleaner SQL and fewer referential errors
    # for users

    op = expr.op()

    if isinstance(op, ops.Selection):
        return _filter_selection(expr, predicates)
    elif isinstance(op, ops.Aggregation):
        # Potential fusion opportunity
        # GH1344: We can't sub in things with correlated subqueries
        simplified_predicates = [
            sub_for(predicate, [(expr, op.table)])
            if not has_reduction(predicate) else predicate
            for predicate in predicates
        ]

        if op.table._is_valid(simplified_predicates):
            result = ops.Aggregation(
                op.table, op.metrics, by=op.by, having=op.having,
                predicates=op.predicates + simplified_predicates,
                sort_keys=op.sort_keys)

            return ir.TableExpr(result)
    elif isinstance(op, ops.Join):
        expr = expr.materialize()

    result = ops.Selection(expr, [], predicates)
    return ir.TableExpr(result)


def _filter_selection(expr, predicates):
    # if any of the filter predicates have the parent expression among
    # their roots, then pushdown (at least of that predicate) is not
    # possible

    # It's not unusual for the filter to reference the projection
    # itself. If a predicate can be pushed down, in this case we must
    # rewrite replacing the table refs with the roots internal to the
    # projection we are referencing
    #
    # Assuming that the fields referenced by the filter predicate originate
    # below the projection, we need to rewrite the predicate referencing
    # the parent tables in the join being projected

    op = expr.op()
    if not op.blocks():
        # Potential fusion opportunity. The predicates may need to be
        # rewritten in terms of the child table. This prevents the broken
        # ref issue (described in more detail in #59)
        simplified_predicates = [
            sub_for(predicate, [(expr, op.table)])
            if not has_reduction(predicate) else predicate
            for predicate in predicates
        ]

        if op.table._is_valid(simplified_predicates):
            result = ops.Selection(
                op.table, [],
                predicates=op.predicates + simplified_predicates,
                sort_keys=op.sort_keys)
            return result.to_expr()

    can_pushdown = _can_pushdown(op, predicates)

    if can_pushdown:
        simplified_predicates = [substitute_parents(x) for x in predicates]
        fused_predicates = op.predicates + simplified_predicates
        result = ops.Selection(op.table,
                               proj_exprs=op.selections,
                               predicates=fused_predicates,
                               sort_keys=op.sort_keys)
    else:
        result = ops.Selection(expr, proj_exprs=[],
                               predicates=predicates)

    return result.to_expr()


def _can_pushdown(op, predicates):
    # Per issues discussed in #173
    #
    # The only case in which pushdown is possible is that all table columns
    # referenced must meet all of the following (not that onerous in practice)
    # criteria
    #
    # 1) Is a table column, not any other kind of expression
    # 2) Is unaliased. So, if you project t3.foo AS bar, then filter on bar,
    #    this cannot be pushed down (until we implement alias rewriting if
    #    necessary)
    # 3) Appears in the selections in the projection (either is part of one of
    #    the entire tables or a single column selection)

    for pred in predicates:
        validator = _PushdownValidate(op, pred)
        predicate_is_valid = validator.get_result()
        if not predicate_is_valid:
            return False
    return True


class _PushdownValidate(object):

    def __init__(self, parent, predicate):
        self.parent = parent
        self.pred = predicate
        self.validator = ExprValidator([self.parent.table])

    def get_result(self):
        predicate = self.pred
        return not has_reduction(predicate) and all(self._walk(predicate))

    def _walk(self, expr):
        def validate(expr):
            op = expr.op()
            if isinstance(op, ops.TableColumn):
                return lin.proceed, self._validate_column(expr)
            return lin.proceed, None

        return lin.traverse(validate, expr, type=ir.ValueExpr)

    def _validate_column(self, expr):
        if isinstance(self.parent, ops.Selection):
            return self._validate_projection(expr)
        else:
            validator = ExprValidator([self.parent.table])
            return validator.validate(expr)

    def _validate_projection(self, expr):
        is_valid = False
        node = expr.op()

        # Has a different alias, invalid
        if _is_aliased(expr):
            return False

        for val in self.parent.selections:
            if (isinstance(val.op(), ops.PhysicalTable) and
                    node.name in val.schema()):
                is_valid = True
            elif (isinstance(val.op(), ops.TableColumn) and
                  node.name == val.get_name() and
                  not _is_aliased(val)):
                # Aliased table columns are no good
                col_table = val.op().table.op()

                lifted_node = substitute_parents(expr).op()

                is_valid = (col_table.is_ancestor(node.table) or
                            col_table.is_ancestor(lifted_node.table))

        return is_valid


def _is_aliased(col_expr):
    return col_expr.op().name != col_expr.get_name()


def windowize_function(expr, w=None):
    def _windowize(x, w):
        if not isinstance(x.op(), ops.WindowOp):
            walked = _walk(x, w)
        else:
            window_arg, window_w = x.op().args
            walked_child = _walk(window_arg, w)

            if walked_child is not window_arg:
                walked = x._factory(ops.WindowOp(walked_child, window_w),
                                    name=x._name)
            else:
                walked = x

        op = walked.op()
        if isinstance(op, ops.AnalyticOp) or getattr(op, '_reduction', False):
            if w is None:
                w = window()
            return walked.over(w)
        elif isinstance(op, ops.WindowOp):
            if w is not None:
                return walked.over(w)
            else:
                return walked
        else:
            return walked

    def _walk(x, w):
        op = x.op()

        unchanged = True
        windowed_args = []
        for arg in op.args:
            if not isinstance(arg, ir.ValueExpr):
                windowed_args.append(arg)
                continue

            new_arg = _windowize(arg, w)
            unchanged = unchanged and arg is new_arg
            windowed_args.append(new_arg)

        if not unchanged:
            new_op = type(op)(*windowed_args)
            return x._factory(new_op, name=x._name)
        else:
            return x

    return _windowize(expr, w)


class Projector(object):

    """
    Analysis and validation of projection operation, taking advantage of
    "projection fusion" opportunities where they exist, i.e. combining
    compatible projections together rather than nesting them. Translation /
    evaluation later will not attempt to do any further fusion /
    simplification.
    """

    def __init__(self, parent, proj_exprs):
        self.parent = parent
        self.input_exprs = proj_exprs
        self.resolved_exprs = [parent._ensure_expr(e) for e in proj_exprs]

        node = self.parent.op()

        if isinstance(node, ops.Selection):
            roots = [node]
        else:
            roots = node.root_tables()

        self.parent_roots = roots

        clean_exprs = []

        for expr in self.resolved_exprs:
            # Perform substitution only if we share common roots
            expr = windowize_function(expr)
            clean_exprs.append(expr)

        self.clean_exprs = clean_exprs

    def get_result(self):
        roots = self.parent_roots
        first_root = roots[0]

        if len(roots) == 1 and isinstance(first_root, ops.Selection):
            fused_op = self._check_fusion(first_root)
            if fused_op is not None:
                return fused_op

        return ops.Selection(self.parent, self.clean_exprs)

    def _check_fusion(self, root):
        roots = root.table._root_tables()
        validator = ExprValidator([root.table])
        fused_exprs = []
        can_fuse = False

        resolved = _maybe_resolve_exprs(root.table, self.input_exprs)
        if not resolved:
            return None

        for val in resolved:
            # XXX
            lifted_val = substitute_parents(val)

            # a * projection
            if (isinstance(val, ir.TableExpr) and
                (self.parent.op().is_ancestor(val) or
                 # gross we share the same table root. Better way to
                 # detect?
                 len(roots) == 1 and val._root_tables()[0] is roots[0])):
                can_fuse = True

                have_root = False
                for y in root.selections:
                    # Don't add the * projection twice
                    if y.equals(root.table):
                        fused_exprs.append(root.table)
                        have_root = True
                        continue
                    fused_exprs.append(y)

                # This was a filter, so implicitly a select *
                if not have_root and len(root.selections) == 0:
                    fused_exprs = [root.table] + fused_exprs
            elif validator.validate(lifted_val):
                can_fuse = True
                fused_exprs.append(lifted_val)
            elif not validator.validate(val):
                can_fuse = False
                break
            else:
                fused_exprs.append(val)

        if can_fuse:
            return ops.Selection(root.table, fused_exprs,
                                 predicates=root.predicates,
                                 sort_keys=root.sort_keys)
        else:
            return None


def _maybe_resolve_exprs(table, exprs):
    try:
        return table._resolve(exprs)
    except (AttributeError, IbisTypeError):
        return None


class ExprValidator(object):

    def __init__(self, exprs):
        self.parent_exprs = exprs

        self.roots = []
        for expr in self.parent_exprs:
            self.roots.extend(expr._root_tables())

    def has_common_roots(self, expr):
        return self.validate(expr)

    def validate(self, expr):
        op = expr.op()
        if isinstance(op, ops.TableColumn):
            if self._among_roots(op.table.op()):
                return True
        elif isinstance(op, ops.Selection):
            if self._among_roots(op):
                return True

        expr_roots = expr._root_tables()
        for root in expr_roots:
            if not self._among_roots(root):
                return False
        return True

    def _among_roots(self, node):
        return self.roots_shared(node) > 0

    def roots_shared(self, node):
        return sum(root.is_ancestor(node) for root in self.roots)

    def shares_some_roots(self, expr):
        expr_roots = expr._root_tables()
        return any(self._among_roots(root)
                   for root in expr_roots)

    def shares_one_root(self, expr):
        expr_roots = expr._root_tables()
        total = sum(self.roots_shared(root)
                    for root in expr_roots)
        return total == 1

    def shares_multiple_roots(self, expr):
        expr_roots = expr._root_tables()
        total = sum(self.roots_shared(expr_roots)
                    for root in expr_roots)
        return total > 1

    def validate_all(self, exprs):
        for expr in exprs:
            self.assert_valid(expr)

    def assert_valid(self, expr):
        if not self.validate(expr):
            msg = self._error_message(expr)
            raise RelationError(msg)

    def _error_message(self, expr):
        return ('The expression %s does not fully originate from '
                'dependencies of the table expression.' % repr(expr))


def fully_originate_from(exprs, parents):
    def finder(expr):
        op = expr.op()

        if isinstance(expr, ir.TableExpr):
            return lin.proceed, expr
        elif op.blocks():
            return lin.halt, None
        else:
            return lin.proceed, None

    # unique table dependencies of exprs and parents
    exprs_deps = set(lin.traverse(finder, exprs))
    parents_deps = set(lin.traverse(finder, parents))

    return exprs_deps <= parents_deps


class FilterValidator(ExprValidator):

    """
    Filters need not necessarily originate fully from the ancestors of the
    table being filtered. The key cases for this are

    - Scalar reductions involving some other tables
    - Array expressions involving other tables only (mapping to "uncorrelated
      subqueries" in SQL-land)
    - Reductions or array expressions like the above, but containing some
      predicate with a record-specific interdependency ("correlated subqueries"
      in SQL)
    """

    def validate(self, expr):
        op = expr.op()

        if (isinstance(expr, ir.BooleanColumn) and
                isinstance(op, ops.TableColumn)):
            return True

        is_valid = True

        if isinstance(op, ops.Contains):
            value_valid = super(FilterValidator, self).validate(op.value)
            is_valid = value_valid
        else:
            roots_valid = []
            for arg in op.flat_args():
                if isinstance(arg, ir.ScalarExpr):
                    # arg_valid = True
                    pass
                elif isinstance(arg, ops.TopKExpr):
                    # TopK not subjected to further analysis for now
                    roots_valid.append(True)
                elif isinstance(arg, (ir.ColumnExpr, ir.AnalyticExpr)):
                    roots_valid.append(self.shares_some_roots(arg))
                elif isinstance(arg, ir.Expr):
                    raise NotImplementedError(repr((type(expr), type(arg))))
                else:
                    # arg_valid = True
                    pass

            is_valid = any(roots_valid)

        return is_valid


def find_source_table(expr):
    """Find the first table expression observed for each argument that the
    expression depends on

    Parameters
    ----------
    expr : ir.Expr

    Returns
    -------
    table_expr : ir.TableExpr

    Examples
    --------
    >>> import ibis
    >>> t = ibis.table([('a', 'double'), ('b', 'string')], name='t')
    >>> expr = t.mutate(c=t.a + 42.0)
    >>> expr  # doctest: +NORMALIZE_WHITESPACE
    ref_0
    UnboundTable[table]
      name: t
      schema:
        a : double
        b : string
    Selection[table]
      table:
        Table: ref_0
      selections:
        Table: ref_0
        c = Add[double*]
          left:
            a = Column[double*] 'a' from table
              ref_0
          right:
            Literal[double]
              42.0
    >>> find_source_table(expr)
    UnboundTable[table]
      name: t
      schema:
        a : double
        b : string
    >>> left = ibis.table([('a', 'int64'), ('b', 'string')])
    >>> right = ibis.table([('c', 'int64'), ('d', 'string')])
    >>> result = left.inner_join(right, left.a == right.c)
    >>> find_source_table(result)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    NotImplementedError: More than one base table not implemented
    """
    def finder(expr):
        if isinstance(expr, ir.TableExpr):
            return lin.halt, expr
        else:
            return lin.proceed, None

    first_tables = lin.traverse(finder, expr.op().flat_args())
    options = list(toolz.unique(first_tables, key=id))

    if len(options) > 1:
        raise NotImplementedError('More than one base table not implemented')

    return options[0]


def flatten_predicate(expr):
    """Yield the expressions corresponding to the `And` nodes of a predicate.

    Parameters
    ----------
    expr : ir.BooleanColumn

    Returns
    -------
    exprs : List[ir.BooleanColumn]

    Examples
    --------
    >>> import ibis
    >>> t = ibis.table([('a', 'int64'), ('b', 'string')], name='t')
    >>> filt = (t.a == 1) & (t.b == 'foo')
    >>> predicates = flatten_predicate(filt)
    >>> len(predicates)
    2
    >>> predicates[0]  # doctest: +NORMALIZE_WHITESPACE
    ref_0
    UnboundTable[table]
      name: t
      schema:
        a : int64
        b : string
    Equals[boolean*]
      left:
        a = Column[int64*] 'a' from table
          ref_0
      right:
        Literal[int64]
          1
    >>> predicates[1]  # doctest: +NORMALIZE_WHITESPACE
    ref_0
    UnboundTable[table]
      name: t
      schema:
        a : int64
        b : string
    Equals[boolean*]
      left:
        b = Column[string*] 'b' from table
          ref_0
      right:
        Literal[string]
          foo
    """
    def predicate(expr):
        if isinstance(expr.op(), ops.And):
            return lin.proceed, None
        else:
            return lin.halt, expr

    return list(lin.traverse(predicate, expr, type=ir.BooleanColumn))



import ibis
import ibis.common as com

import numpy as np

import ibis.expr.datatypes as dt
import ibis.expr.types as ir

import ibis.sql.compiler as comp
import ibis.expr.operations as ops
from ibis.impala.compiler import ImpalaSelect, unary, fixed_arity
from ibis.impala import compiler as impala_compiler


class BigQuerySelectBuilder(comp.SelectBuilder):

    @property
    def _select_class(self):
        return BigQuerySelect


class BigQueryQueryBuilder(comp.QueryBuilder):

    select_builder = BigQuerySelectBuilder

    @property
    def _union_class(self):
        # return BigQueryUnion
        raise NotImplementedError()


def build_ast(expr, context):
    builder = BigQueryQueryBuilder(expr, context=context)
    return builder.get_result()


def _get_query(expr, context):
    ast = build_ast(expr, context)
    (query, rest) = (ast.queries[0], ast.queries[1:])
    assert not rest
    return query


def to_sql(expr, context):
    query = _get_query(expr, context)
    compiled = query.compile()
    return compiled


class BigQueryContext(comp.QueryContext):

    def _to_sql(self, expr, ctx):
        return to_sql(expr, context=ctx)


def _extract_field(sql_attr):
    def extract_field_formatter(translator, expr):
        op = expr.op()
        arg = translator.translate(op.args[0])
        return 'EXTRACT({} from {})'.format(sql_attr, arg)
    return extract_field_formatter


_sql_type_names = {
    'int8': 'int64',
    'int16': 'int64',
    'int32': 'int64',
    'int64': 'int64',
    'float': 'float64',
    'double': 'float64',
    'string': 'string',
    'boolean': 'boolean',
    'timestamp': 'timestamp',
    'date': 'date',
}


def _cast(translator, expr):
    op = expr.op()
    arg, target_type = op.args
    arg_formatted = translator.translate(arg)
    sql_type = _sql_type_names[target_type.name.lower()]
    return 'CAST({} AS {})'.format(arg_formatted, sql_type.upper())


def _struct_field(translator, expr):
    arg, field = expr.op().args
    arg_formatted = translator.translate(arg)
    return '{}.`{}`'.format(arg_formatted, field)


def _array_concat(translator, expr):
    return 'ARRAY_CONCAT({})'.format(
        ', '.join(map(translator.translate, expr.op().args))
    )


def _array_index(translator, expr):
    # SAFE_OFFSET returns NULL if out of bounds
    return '{}[SAFE_OFFSET({})]'.format(
        *map(translator.translate, expr.op().args)
    )


def _string_find(translator, expr):
    haystack, needle, start, end = expr.op().args

    if start is not None:
        raise NotImplementedError('start not implemented for string find')
    if end is not None:
        raise NotImplementedError('end not implemented for string find')

    return 'STRPOS({}, {}) - 1'.format(
        translator.translate(haystack),
        translator.translate(needle)
    )


def _translate_pattern(translator, pattern):
    # add 'r' to string literals to indicate to BigQuery this is a raw string
    return 'r' * isinstance(pattern.op(), ir.Literal) + translator.translate(
        pattern
    )


def _regex_search(translator, expr):
    arg, pattern = expr.op().args
    regex = _translate_pattern(translator, pattern)
    result = 'REGEXP_CONTAINS({}, {})'.format(translator.translate(arg), regex)
    return result


def _regex_extract(translator, expr):
    arg, pattern, index = expr.op().args
    regex = _translate_pattern(translator, pattern)
    result = 'REGEXP_EXTRACT_ALL({}, {})[SAFE_OFFSET({})]'.format(
        translator.translate(arg),
        regex,
        translator.translate(index)
    )
    return result


def _regex_replace(translator, expr):
    arg, pattern, replacement = expr.op().args
    regex = _translate_pattern(translator, pattern)
    result = 'REGEXP_REPLACE({}, {}, {})'.format(
        translator.translate(arg),
        regex,
        translator.translate(replacement),
    )
    return result


def _string_concat(translator, expr):
    return 'CONCAT({})'.format(
        ', '.join(map(translator.translate, expr.op().args))
    )


def _string_join(translator, expr):
    sep, args = expr.op().args
    return 'ARRAY_TO_STRING([{}], {})'.format(
        ', '.join(map(translator.translate, args)),
        translator.translate(sep)
    )


def _string_ascii(translator, expr):
    arg, = expr.op().args
    return 'TO_CODE_POINTS({})[SAFE_OFFSET(0)]'.format(
        translator.translate(arg)
    )


def _string_right(translator, expr):
    arg, nchars = map(translator.translate, expr.op().args)
    return 'SUBSTR({arg}, -LEAST(LENGTH({arg}), {nchars}))'.format(
        arg=arg,
        nchars=nchars,
    )


def _array_literal_format(expr):
    return str(list(expr.op().value))


def _log(translator, expr):
    op = expr.op()
    arg, base = op.args
    arg_formatted = translator.translate(arg)

    if base is None:
        return 'ln({})'.format(arg_formatted)

    base_formatted = translator.translate(base)
    return 'log({}, {})'.format(arg_formatted, base_formatted)


def _literal(translator, expr):

    if isinstance(expr, ir.NumericValue):
        value = expr.op().value
        if not np.isfinite(value):
            return 'CAST({!r} AS FLOAT64)'.format(str(value))

    try:
        return impala_compiler._literal(translator, expr)
    except NotImplementedError:
        if isinstance(expr, ir.ArrayValue):
            return _array_literal_format(expr)
        raise NotImplementedError(type(expr).__name__)


def _arbitrary(translator, expr):
    arg, how, where = expr.op().args

    if where is not None:
        arg = where.ifelse(arg, ibis.NA)

    if how != 'first':
        raise com.UnsupportedOperationError(
            '{!r} value not supported for arbitrary in BigQuery'.format(how)
        )

    return 'ANY_VALUE({})'.format(translator.translate(arg))


_date_units = {
    'Y': 'YEAR',
    'Q': 'QUARTER',
    'W': 'WEEK',
    'M': 'MONTH',
}


_timestamp_units = {
    'us': 'MICROSECOND',
    'ms': 'MILLISECOND',
    's': 'SECOND',
    'm': 'MINUTE',
    'h': 'HOUR',
}
_timestamp_units.update(_date_units)


def _truncate(kind, units):
    def truncator(translator, expr):
        op = expr.op()
        arg, unit = op.args

        arg = translator.translate(op.args[0])
        try:
            unit = units[unit]
        except KeyError:
            raise com.UnsupportedOperationError(
                '{!r} unit is not supported in timestamp truncate'.format(unit)
            )

        return "{}_TRUNC({}, {})".format(kind, arg, unit)
    return truncator


def _timestamp_op(func, units):
    def _formatter(translator, expr):
        op = expr.op()
        arg, offset = op.args

        if offset.unit not in units:
            raise com.UnsupportedOperationError(
                'BigQuery does not allow binary operation '
                '{} with INTERVAL offset {}'.format(func, offset.unit)
            )
        formatted_arg = translator.translate(arg)
        formatted_offset = translator.translate(offset)
        result = '{}({}, {})'.format(func, formatted_arg, formatted_offset)
        return result

    return _formatter


_operation_registry = impala_compiler._operation_registry.copy()
_operation_registry.update({
    ops.ExtractYear: _extract_field('year'),
    ops.ExtractMonth: _extract_field('month'),
    ops.ExtractDay: _extract_field('day'),
    ops.ExtractHour: _extract_field('hour'),
    ops.ExtractMinute: _extract_field('minute'),
    ops.ExtractSecond: _extract_field('second'),
    ops.ExtractMillisecond: _extract_field('millisecond'),

    ops.StringReplace: fixed_arity('REPLACE', 3),
    ops.StringSplit: fixed_arity('SPLIT', 2),
    ops.StringConcat: _string_concat,
    ops.StringJoin: _string_join,
    ops.StringAscii: _string_ascii,
    ops.StringFind: _string_find,
    ops.StrRight: _string_right,
    ops.Repeat: fixed_arity('REPEAT', 2),
    ops.RegexSearch: _regex_search,
    ops.RegexExtract: _regex_extract,
    ops.RegexReplace: _regex_replace,

    ops.GroupConcat: fixed_arity('STRING_AGG', 2),

    ops.IfNull: fixed_arity('IFNULL', 2),
    ops.Cast: _cast,

    ops.StructField: _struct_field,

    ops.ArrayCollect: unary('ARRAY_AGG'),
    ops.ArrayConcat: _array_concat,
    ops.ArrayIndex: _array_index,
    ops.ArrayLength: unary('ARRAY_LENGTH'),

    ops.Log: _log,
    ops.Modulus: fixed_arity('MOD', 2),

    ops.Date: unary('DATE'),

    # BigQuery doesn't have these operations built in.
    # ops.ArrayRepeat: _array_repeat,
    # ops.ArraySlice: _array_slice,
    ir.Literal: _literal,
    ops.Arbitrary: _arbitrary,

    ops.TimestampTruncate: _truncate('TIMESTAMP', _timestamp_units),
    ops.DateTruncate: _truncate('DATE', _date_units),

    ops.TimestampAdd: _timestamp_op(
        'TIMESTAMP_ADD', {'h', 'm', 's', 'ms', 'us'}),
    ops.TimestampSub: _timestamp_op(
        'TIMESTAMP_DIFF', {'h', 'm', 's', 'ms', 'us'}),

    ops.DateAdd: _timestamp_op('DATE_ADD', {'D', 'W', 'M', 'Q', 'Y'}),
    ops.DateSub: _timestamp_op('DATE_SUB', {'D', 'W', 'M', 'Q', 'Y'}),
})

_invalid_operations = {
    ops.Translate,
    ops.FindInSet,
    ops.Capitalize,
    ops.DateDiff,
    ops.TimestampDiff
}

_operation_registry = {
    k: v for k, v in _operation_registry.items()
    if k not in _invalid_operations
}


class BigQueryExprTranslator(impala_compiler.ImpalaExprTranslator):
    _registry = _operation_registry
    _rewrites = impala_compiler.ImpalaExprTranslator._rewrites.copy()

    context_class = BigQueryContext

    def _trans_param(self, expr):
        op = expr.op()
        if op not in self.context.params:
            raise KeyError(op)
        return '@{}'.format(expr.get_name())


rewrites = BigQueryExprTranslator.rewrites


@rewrites(ops.Any)
def bigquery_rewrite_any(expr):
    arg, = expr.op().args
    return arg.cast(dt.int64).sum() > 0


@rewrites(ops.NotAny)
def bigquery_rewrite_notany(expr):
    arg, = expr.op().args
    return arg.cast(dt.int64).sum() == 0


@rewrites(ops.All)
def bigquery_rewrite_all(expr):
    arg, = expr.op().args
    return (1 - arg.cast(dt.int64)).sum() == 0


@rewrites(ops.NotAll)
def bigquery_rewrite_notall(expr):
    arg, = expr.op().args
    return (1 - arg.cast(dt.int64)).sum() != 0


class BigQuerySelect(ImpalaSelect):

    translator = BigQueryExprTranslator


@rewrites(ops.IdenticalTo)
def identical_to(expr):
    left, right = expr.op().args
    return (left.isnull() & right.isnull()) | (left == right)


@rewrites(ops.Log2)
def log2(expr):
    arg, = expr.op().args
    return arg.log(2)


class BigQueryDialect(impala_compiler.ImpalaDialect):

    translator = BigQueryExprTranslator


dialect = BigQueryDialect



import pytest
import operator
import numpy as np
import pandas as pd
import pandas.util.testing as tm
from datetime import date, datetime

import ibis
import ibis.expr.datatypes as dt
from ibis import literal as L


pytest.importorskip('clickhouse_driver')
pytestmark = pytest.mark.clickhouse


@pytest.mark.parametrize(('left', 'right', 'type'), [
    (L('2017-04-01'), date(2017, 4, 2), dt.date),
    (date(2017, 4, 2), L('2017-04-01'), dt.date),
    (L('2017-04-01 01:02:33'), datetime(2017, 4, 1, 1, 3, 34), dt.timestamp),
    (datetime(2017, 4, 1, 1, 3, 34), L('2017-04-01 01:02:33'), dt.timestamp)
])
@pytest.mark.parametrize('op', [
    operator.eq,
    operator.ne,
    operator.lt,
    operator.le,
    operator.gt,
    operator.ge,
])
def test_string_temporal_compare(con, op, left, right, type):
    expr = op(left, right)
    result = con.execute(expr)
    left_raw = con.execute(L(left).cast(type))
    right_raw = con.execute(L(right).cast(type))
    expected = op(left_raw, right_raw)
    assert result == expected


@pytest.mark.parametrize(('func', 'left', 'right', 'expected'), [
    (operator.add, L(3), L(4), 7),
    (operator.sub, L(3), L(4), -1),
    (operator.mul, L(3), L(4), 12),
    (operator.truediv, L(12), L(4), 3),
    (operator.pow, L(12), L(2), 144),
    (operator.mod, L(12), L(5), 2),
    (operator.truediv, L(7), L(2), 3.5),
    (operator.floordiv, L(7), L(2), 3),
    (lambda x, y: x.floordiv(y), L(7), 2, 3),
    (lambda x, y: x.rfloordiv(y), L(2), 7, 3)
])
def test_binary_arithmetic(con, func, left, right, expected):
    expr = func(left, right)
    result = con.execute(expr)
    assert result == expected


@pytest.mark.parametrize(('op', 'expected'), [
    (lambda a, b: a + b, '`int_col` + `tinyint_col`'),
    (lambda a, b: a - b, '`int_col` - `tinyint_col`'),
    (lambda a, b: a * b, '`int_col` * `tinyint_col`'),
    (lambda a, b: a / b, '`int_col` / `tinyint_col`'),
    (lambda a, b: a ** b, 'pow(`int_col`, `tinyint_col`)'),
    (lambda a, b: a < b, '`int_col` < `tinyint_col`'),
    (lambda a, b: a <= b, '`int_col` <= `tinyint_col`'),
    (lambda a, b: a > b, '`int_col` > `tinyint_col`'),
    (lambda a, b: a >= b, '`int_col` >= `tinyint_col`'),
    (lambda a, b: a == b, '`int_col` = `tinyint_col`'),
    (lambda a, b: a != b, '`int_col` != `tinyint_col`')
])
def test_binary_infix_operators(con, alltypes, translate, op, expected):
    a, b = alltypes.int_col, alltypes.tinyint_col
    expr = op(a, b)
    assert translate(expr) == expected
    assert len(con.execute(expr))


# TODO: test boolean operators
# (h & bool_col, '`h` AND (`a` > 0)'),
# (h | bool_col, '`h` OR (`a` > 0)'),
# (h ^ bool_col, 'xor(`h`, (`a` > 0))')


@pytest.mark.parametrize(('op', 'expected'), [
    (lambda a, b, c: (a + b) + c,
     '(`int_col` + `tinyint_col`) + `double_col`'),
    (lambda a, b, c: a.log() + c,
     'log(`int_col`) + `double_col`'),
    (lambda a, b, c: (b + (-(a + c))),
     '`tinyint_col` + (-(`int_col` + `double_col`))')
])
def test_binary_infix_parenthesization(con, alltypes, translate, op, expected):
    a = alltypes.int_col
    b = alltypes.tinyint_col
    c = alltypes.double_col

    expr = op(a, b, c)
    assert translate(expr) == expected
    assert len(con.execute(expr))


def test_between(con, alltypes, translate):
    expr = alltypes.int_col.between(0, 10)
    assert translate(expr) == '`int_col` BETWEEN 0 AND 10'
    assert len(con.execute(expr))


@pytest.mark.parametrize(('left', 'right'), [
    (L('2017-03-31').cast(dt.date), date(2017, 4, 2)),
    (date(2017, 3, 31), L('2017-04-02').cast(dt.date))
])
def test_string_temporal_compare_between_dates(con, left, right):
    expr = ibis.timestamp('2017-04-01').cast(dt.date).between(left, right)
    result = con.execute(expr)
    assert result


@pytest.mark.parametrize(('left', 'right'), [
    (
        L('2017-03-31 00:02:33').cast(dt.timestamp),
        datetime(2017, 4, 1, 1, 3, 34),
    ),
    (
        datetime(2017, 3, 31, 0, 2, 33),
        L('2017-04-01 01:03:34').cast(dt.timestamp),
    )
])
def test_string_temporal_compare_between_datetimes(con, left, right):
    expr = ibis.timestamp('2017-04-01 00:02:34').between(left, right)
    result = con.execute(expr)
    assert result


def test_field_in_literals(con, alltypes, translate):
    expr = alltypes.string_col.isin(['foo', 'bar', 'baz'])
    assert translate(expr) == "`string_col` IN ('foo', 'bar', 'baz')"
    assert len(con.execute(expr))

    expr = alltypes.string_col.notin(['foo', 'bar', 'baz'])
    assert translate(expr) == "`string_col` NOT IN ('foo', 'bar', 'baz')"
    assert len(con.execute(expr))


@pytest.mark.parametrize('column', [
    'int_col',
    'float_col',
    'bool_col'
])
def test_negate(con, alltypes, translate, column):
    # clickhouse represent boolean as UInt8
    expr = -getattr(alltypes, column)
    assert translate(expr) == '-`{0}`'.format(column)
    assert len(con.execute(expr))


@pytest.mark.parametrize('field', [
    'tinyint_col',
    'smallint_col',
    'int_col',
    'bigint_col',
    'float_col',
    'double_col',
    'year',
    'month',
])
def test_negate_non_boolean(con, alltypes, field, df):
    t = alltypes.limit(10)
    expr = t.projection([(-t[field]).name(field)])
    result = expr.execute()[field]
    expected = -df.head(10)[field]
    tm.assert_series_equal(result, expected)


def test_negate_literal(con):
    expr = -L(5.245)
    assert round(con.execute(expr), 3) == -5.245


@pytest.mark.parametrize(('op', 'pandas_op'), [
    (
        lambda t: (t.double_col > 20).ifelse(10, -20),
        lambda df: pd.Series(np.where(df.double_col > 20, 10, -20),
                             dtype='int8')
    ),
    (
        lambda t: (t.double_col > 20).ifelse(10, -20).abs(),
        lambda df: (pd.Series(np.where(df.double_col > 20, 10, -20))
                    .abs()
                    .astype('int8'))
    ),
])
def test_ifelse(alltypes, df, op, pandas_op, translate):
    expr = op(alltypes)
    result = expr.execute()
    result.name = None
    expected = pandas_op(df)

    tm.assert_series_equal(result, expected)


def test_simple_case(con, alltypes, translate):
    t = alltypes
    expr = (t.string_col.case()
            .when('foo', 'bar')
            .when('baz', 'qux')
            .else_('default')
            .end())

    expected = """CASE `string_col`
  WHEN 'foo' THEN 'bar'
  WHEN 'baz' THEN 'qux'
  ELSE 'default'
END"""
    assert translate(expr) == expected
    assert len(con.execute(expr))


def test_search_case(con, alltypes, translate):
    t = alltypes
    expr = (ibis.case()
            .when(t.float_col > 0, t.int_col * 2)
            .when(t.float_col < 0, t.int_col)
            .else_(0)
            .end())

    expected = """CASE
  WHEN `float_col` > 0 THEN `int_col` * 2
  WHEN `float_col` < 0 THEN `int_col`
  ELSE 0
END"""
    assert translate(expr) == expected
    assert len(con.execute(expr))


# TODO: Clickhouse raises incompatible type error
# def test_bucket_to_case(con, alltypes, translate):
#     buckets = [0, 10, 25, 50]

#     expr1 = alltypes.float_col.bucket(buckets)
#     expected1 = """\
# CASE
#   WHEN (`float_col` >= 0) AND (`float_col` < 10) THEN 0
#   WHEN (`float_col` >= 10) AND (`float_col` < 25) THEN 1
#   WHEN (`float_col` >= 25) AND (`float_col` <= 50) THEN 2
#   ELSE Null
# END"""

#     assert translate(expr1) == expected1
#     assert len(con.execute(expr1))



import os

import pytest

pytest.importorskip('graphviz')

import ibis  # noqa: E402
import ibis.expr.types as ir  # noqa: E402
import ibis.expr.visualize as viz  # noqa: E402

from ibis.expr import rules  # noqa: E402

pytestmark = pytest.mark.skipif(
    int(os.environ.get('CONDA_BUILD', 0)) == 1, reason='CONDA_BUILD defined'
)


@pytest.mark.parametrize(
    'expr_func',
    [
        lambda t: t.a,
        lambda t: t.a + t.b,
        lambda t: t.a + t.b > 3 ** t.a,
        lambda t: t[(t.a + t.b * 2 * t.b / t.b ** 3 > 4) & (t.b > 5)],
        lambda t: t[(t.a + t.b * 2 * t.b / t.b ** 3 > 4) & (t.b > 5)].group_by(
            'c'
        ).aggregate(
            amean=lambda f: f.a.mean(),
            bsum=lambda f: f.b.sum(),
        )
    ]
)
def test_exprs(table, expr_func):
    expr = expr_func(table)
    graph = viz.to_graph(expr)
    assert str(hash(repr(table.op()))) in graph.source
    assert str(hash(repr(expr.op()))) in graph.source


def test_custom_expr():
    class MyExpr(ir.Expr):
        pass

    class MyExprNode(ir.Node):

        input_type = [
            rules.string(name='foo'),
            rules.number(name='bar'),
        ]

        def output_type(self):
            return MyExpr

    op = MyExprNode(['Hello!', 42.3])
    expr = op.to_expr()
    graph = viz.to_graph(expr)
    assert str(hash(repr(op))) in graph.source


def test_custom_expr_with_not_implemented_type():

    class MyExpr(ir.Expr):
        def type(self):
            raise NotImplementedError

        def schema(self):
            raise NotImplementedError

    class MyExprNode(ir.Node):

        input_type = [
            rules.string(name='foo'),
            rules.number(name='bar'),
        ]

        def output_type(self):
            return MyExpr

    op = MyExprNode(['Hello!', 42.3])
    expr = op.to_expr()
    graph = viz.to_graph(expr)
    assert str(hash(repr(op))) in graph.source


@pytest.mark.parametrize('how', ['inner', 'left', 'right', 'outer'])
def test_join(how):
    left = ibis.table([('a', 'int64'), ('b', 'string')])
    right = ibis.table([('b', 'string'), ('c', 'int64')])
    joined = left.join(right, left.b == right.b, how=how)
    result = joined[left.a, right.c]
    graph = viz.to_graph(result)
    assert str(hash(repr(result.op()))) in graph.source


def test_sort_by():
    t = ibis.table([('a', 'int64'), ('b', 'string'), ('c', 'int32')])
    expr = t.groupby(t.b).aggregate(
        sum_a=t.a.sum().cast('double')
    ).sort_by('c')
    graph = viz.to_graph(expr)
    assert str(hash(repr(expr.op()))) in graph.source


@pytest.mark.skipif(
    bool(os.environ.get('APPVEYOR', None)),
    reason='Not sure what the prerequisites for running this on Windows are'
)
def test_optional_graphviz_repr():
    t = ibis.table([('a', 'int64'), ('b', 'string'), ('c', 'int32')])
    expr = t.groupby(t.b).aggregate(
        sum_a=t.a.sum().cast('double')
    ).sort_by('c')

    # default behavior
    assert expr._repr_png_() is not None

    # turn it off
    ibis.options.graphviz_repr = False
    assert expr._repr_png_() is None

    # turn it back on
    ibis.options.graphviz_repr = True
    assert expr._repr_png_() is not None



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

import ibis
import ibis.common as com

from ibis import window
from ibis.tests.util import assert_equal

pytest.importorskip('hdfs')
pytest.importorskip('sqlalchemy')
pytest.importorskip('impala.dbapi')

from ibis.impala.compiler import to_sql  # noqa: E402
from ibis.impala.tests.common import ImpalaE2E  # noqa: E402


@pytest.yield_fixture(scope='module')
def con(request):
    ImpalaE2E.setUpClass()

    try:
        yield ImpalaE2E.con
    finally:
        ImpalaE2E.tearDownClass()


def assert_sql_equal(expr, expected):
    result = to_sql(expr)
    assert result == expected


@pytest.mark.impala
def test_aggregate_in_projection(con):
    t = con.table('alltypes')
    proj = t[t, (t.f / t.f.sum()).name('normed_f')]

    expected = """\
SELECT *, `f` / sum(`f`) OVER () AS `normed_f`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(proj, expected)


@pytest.mark.impala
def test_add_default_order_by(con):
    t = con.table('alltypes')

    first = t.f.first().name('first')
    last = t.f.last().name('last')
    lag = t.f.lag().name('lag')
    diff = (t.f.lead() - t.f).name('fwd_diff')
    lag2 = t.f.lag().over(window(order_by=t.d)).name('lag2')
    grouped = t.group_by('g')
    proj = grouped.mutate([lag, diff, first, last, lag2])
    expected = """\
SELECT *, lag(`f`) OVER (PARTITION BY `g` ORDER BY `f`) AS `lag`,
       lead(`f`) OVER (PARTITION BY `g` ORDER BY `f`) - `f` AS `fwd_diff`,
       first_value(`f`) OVER (PARTITION BY `g` ORDER BY `f`) AS `first`,
       last_value(`f`) OVER (PARTITION BY `g` ORDER BY `f`) AS `last`,
       lag(`f`) OVER (PARTITION BY `g` ORDER BY `d`) AS `lag2`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(proj, expected)


@pytest.mark.impala
@pytest.mark.parametrize(
    ['window', 'frame'],
    [
        (window(preceding=0),
         'range between current row and unbounded following'),

        (window(following=0),
         'range between unbounded preceding and current row'),

        (window(preceding=5),
         'rows between 5 preceding and unbounded following'),
        (window(preceding=5, following=0),
         'rows between 5 preceding and current row'),
        (window(preceding=5, following=2),
         'rows between 5 preceding and 2 following'),
        (window(following=2),
         'rows between unbounded preceding and 2 following'),
        (window(following=2, preceding=0),
         'rows between current row and 2 following'),
        (window(preceding=5),
         'rows between 5 preceding and unbounded following'),
        (window(following=[5, 10]),
         'rows between 5 following and 10 following'),
        (window(preceding=[10, 5]),
         'rows between 10 preceding and 5 preceding'),

        # # cumulative windows
        (ibis.cumulative_window(),
         'range between unbounded preceding and current row'),

        # # trailing windows
        (ibis.trailing_window(10),
         'rows between 10 preceding and current row'),
    ]
)
def test_window_frame_specs(con, window, frame):
    t = con.table('alltypes')

    ex_template = """\
SELECT sum(`d`) OVER (ORDER BY `f` {0}) AS `foo`
FROM ibis_testing.`alltypes`"""

    w2 = window.order_by(t.f)
    expr = t.projection([t.d.sum().over(w2).name('foo')])
    expected = ex_template.format(frame.upper())
    assert_sql_equal(expr, expected)


@pytest.mark.impala
def test_cumulative_functions(con):
    t = con.table('alltypes')

    w = ibis.window(order_by=t.d)
    exprs = [
        (t.f.cumsum().over(w), t.f.sum().over(w)),
        (t.f.cummin().over(w), t.f.min().over(w)),
        (t.f.cummax().over(w), t.f.max().over(w)),
        (t.f.cummean().over(w), t.f.mean().over(w)),
    ]

    for cumulative, static in exprs:
        actual = cumulative.name('foo')
        expected = static.over(ibis.cumulative_window()).name('foo')

        expr1 = t.projection(actual)
        expr2 = t.projection(expected)

        assert to_sql(expr1) == to_sql(expr2)


@pytest.mark.impala
def test_nested_analytic_function(con):
    t = con.table('alltypes')

    w = window(order_by=t.f)
    expr = (t.f - t.f.lag()).lag().over(w).name('foo')
    result = t.projection([expr])
    expected = """\
SELECT lag(`f` - lag(`f`) OVER (ORDER BY `f`)) \
OVER (ORDER BY `f`) AS `foo`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(result, expected)


@pytest.mark.impala
def test_rank_functions(con):
    t = con.table('alltypes')

    proj = t[t.g, t.f.rank().name('minr'),
             t.f.dense_rank().name('denser')]
    expected = """\
SELECT `g`, (rank() OVER (ORDER BY `f`) - 1) AS `minr`,
       (dense_rank() OVER (ORDER BY `f`) - 1) AS `denser`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(proj, expected)


@pytest.mark.impala
def test_multiple_windows(con):
    t = con.table('alltypes')

    w = window(group_by=t.g)

    expr = t.f.sum().over(w) - t.f.sum()
    proj = t.projection([t.g, expr.name('result')])

    expected = """\
SELECT `g`, sum(`f`) OVER (PARTITION BY `g`) - sum(`f`) OVER () AS `result`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(proj, expected)


@pytest.mark.impala
def test_order_by_desc(con):
    t = con.table('alltypes')

    w = window(order_by=ibis.desc(t.f))

    proj = t[t.f, ibis.row_number().over(w).name('revrank')]
    expected = """\
SELECT `f`, (row_number() OVER (ORDER BY `f` DESC) - 1) AS `revrank`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(proj, expected)

    expr = (t.group_by('g')
            .order_by(ibis.desc(t.f))
            [t.d.lag().name('foo'), t.a.max()])
    expected = """\
SELECT lag(`d`) OVER (PARTITION BY `g` ORDER BY `f` DESC) AS `foo`,
       max(`a`) OVER (PARTITION BY `g` ORDER BY `f` DESC) AS `max`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(expr, expected)


@pytest.mark.impala
def test_row_number_does_not_require_order_by(con):
    t = con.table('alltypes')

    expr = t.group_by(t.g).mutate(ibis.row_number().name('foo'))
    expected = """\
SELECT *, (row_number() OVER (PARTITION BY `g`) - 1) AS `foo`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(expr, expected)

    expr = (t.group_by(t.g)
            .order_by(t.f)
            .mutate(ibis.row_number().name('foo')))

    expected = """\
SELECT *, (row_number() OVER (PARTITION BY `g` ORDER BY `f`) - 1) AS `foo`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(expr, expected)


@pytest.mark.impala
def test_row_number_properly_composes_with_arithmetic(con):
    t = con.table('alltypes')
    w = ibis.window(order_by=t.f)
    expr = t.mutate(new=ibis.row_number().over(w) / 2)

    expected = """\
SELECT *, (row_number() OVER (ORDER BY `f`) - 1) / 2 AS `new`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(expr, expected)


@pytest.mark.impala
@pytest.mark.parametrize(
    ['column', 'op'],
    [
        ('f', 'approx_nunique'),
        ('f', 'approx_median'),
        ('g', 'group_concat'),
    ]
)
def test_unsupported_aggregate_functions(con, column, op):
    t = con.table('alltypes')
    w = ibis.window(order_by=t.d)
    expr = getattr(t[column], op)()
    proj = t.projection([expr.over(w).name('foo')])
    with pytest.raises(com.TranslationError):
        to_sql(proj)


@pytest.mark.impala
def test_propagate_nested_windows(con):
    # GH #469
    t = con.table('alltypes')

    w = ibis.window(group_by=t.g, order_by=t.f)

    col = (t.f - t.f.lag()).lag()

    # propagate down here!
    result = col.over(w)
    ex_expr = (t.f - t.f.lag().over(w)).lag().over(w)
    assert_equal(result, ex_expr)

    expr = t.projection(col.over(w).name('foo'))
    expected = """\
SELECT lag(`f` - lag(`f`) OVER (PARTITION BY `g` ORDER BY `f`)) \
OVER (PARTITION BY `g` ORDER BY `f`) AS `foo`
FROM ibis_testing.`alltypes`"""
    assert_sql_equal(expr, expected)


@pytest.mark.xfail
def test_math_on_windowed_expr():
    # Window clause may not be found at top level of expression
    assert False


@pytest.mark.xfail
def test_group_by_then_different_sort_orders():
    assert False



from __future__ import absolute_import

import ibis
import toolz

from ibis.pandas.client import PandasClient
from ibis.pandas.decimal import execute_node  # noqa: F401
from ibis.pandas.execution import execute  # noqa: F401


__all__ = 'connect', 'execute', 'dialect'


def connect(dictionary):
    """Construct a pandas client from a dictionary of DataFrames.

    Parameters
    ----------
    dictionary : dict

    Returns
    -------
    PandasClient
    """
    return PandasClient(dictionary)


def from_dataframe(df, name='df', client=None):
    """
    convenience function to construct an ibis table
    from a DataFrame

    EXPERIMENTAL API

    Parameters
    ----------
    df : DataFrame
    name : str, default 'df'
    client : Client, default new PandasClient
        client dictionary will be mutated with the
        name of the DataFrame

    Returns
    -------
    Table
    """

    if client is None:
        return connect({name: df}).table(name)
    client.dictionary[name] = df
    return client.table(name)


def _flatten_subclass_tree(cls):
    """Return the set of all child classes of `cls`.

    Parameters
    ----------
    cls : Type

    Returns
    -------
    frozenset[Type]
    """
    subclasses = frozenset(cls.__subclasses__())
    children = frozenset(toolz.concat(map(_flatten_subclass_tree, subclasses)))
    return frozenset({cls}) | subclasses | children


class PandasExprTranslator(object):
    # get the dispatched functions from the execute_node dispatcher and compute
    # and flatten the type tree of the first argument which is always the Node
    # subclass
    _registry = frozenset(toolz.concat(
        _flatten_subclass_tree(types[0]) for types in execute_node.funcs
    ))
    _rewrites = {}


class PandasDialect(ibis.sql.compiler.Dialect):

    translator = PandasExprTranslator


PandasClient.dialect = dialect = PandasDialect



from __future__ import absolute_import

import itertools

import numpy as np

import sys
import six
from six import BytesIO, StringIO, string_types  # noqa: F401


PY2 = sys.version_info[0] == 2


if not PY2:
    unicode_type = str

    def lzip(*x):
        return list(zip(*x))

    zip = zip
    zip_longest = itertools.zip_longest

    def viewkeys(x):
        return x.keys()

    from decimal import Decimal
    from inspect import signature, Parameter, _empty
    import unittest.mock as mock
    range = range
    import builtins
    import pickle
    maketrans = str.maketrans
    import functools
    from functools import reduce
else:
    try:
        from cdecimal import Decimal
    except ImportError:
        from decimal import Decimal  # noqa: F401

    from funcsigs import signature, Parameter, _empty  # noqa: F401

    unicode_type = unicode  # noqa: F821
    lzip = zip
    zip = itertools.izip
    zip_longest = itertools.izip_longest

    def viewkeys(x):
        return x.viewkeys()

    try:
        import mock  # noqa: F401
    except ImportError:
        pass

    import __builtin__ as builtins  # noqa: F401
    import functools32 as functools  # noqa: F401

    range = xrange  # noqa: F821
    reduce = reduce  # noqa: F821
    import cPickle as pickle  # noqa: F401

    def maketrans(x, y):
        import string

        assert type(x) == type(y), 'type(x) != type(y) -> {} != {}'.format(
            type(x), type(y)
        )
        assert len(x) == len(y), 'len(x) != len(y) -> {:d} != {:d}'.format(
            len(x), len(y)
        )

        if isinstance(x, six.text_type):
            return dict(zip(map(ord, x), y))
        return string.maketrans(x, y)


integer_types = six.integer_types + (np.integer,)


try:
    from pathlib import Path  # noqa: F401
except ImportError:
    # py2 compat
    from pathlib2 import Path  # noqa: F401


try:
    from contextlib import suppress
except ImportError:
    from contextlib import contextmanager

    @contextmanager
    def suppress(*exceptions):
        try:
            yield
        except exceptions:
            pass


try:
    from pkg_resources import parse_version  # noqa: F401
except ImportError:
    from distutils.version import LooseVersion as parse_version  # noqa: F401


# pandas compat
try:
    from pandas.api.types import (DatetimeTZDtype,  # noqa: F401
                                  CategoricalDtype)  # noqa: F401
except ImportError:
    from pandas.types.dtypes import (DatetimeTZDtype,  # noqa: F401
                                     CategoricalDtype)  # noqa: F401

try:
    from pandas.core.tools.datetimes import to_time, to_datetime  # noqa: F401
except ImportError:
    from pandas.tseries.tools import to_time, to_datetime  # noqa: F401


def to_date(*args, **kwargs):
    return to_datetime(*args, **kwargs).date()



from ibis.pandas.execution.generic import *  # noqa: F401,F403
from ibis.pandas.execution.strings import *  # noqa: F401,F403
from ibis.pandas.execution.temporal import *  # noqa: F401,F403
from ibis.pandas.execution.arrays import *  # noqa: F401,F403
from ibis.pandas.execution.selection import *  # noqa: F401,F403
from ibis.pandas.execution.join import *  # noqa: F401,F403
from ibis.pandas.execution.window import *  # noqa: F401,F403



import pytest
import datetime
from operator import methodcaller

import numpy as np
import pandas as pd
import pandas.util.testing as tm  # noqa: E402
import ibis
from ibis.expr import datatypes as dt
from ibis import literal as L  # noqa: E402
from ibis.compat import PY2


pytestmark = pytest.mark.pandas


@pytest.mark.parametrize(
    ('case_func', 'expected_func'),
    [
        (lambda v: v.strftime('%Y%m%d'), lambda vt: vt.strftime('%Y%m%d')),

        (lambda v: v.year(), lambda vt: vt.year),
        (lambda v: v.month(), lambda vt: vt.month),
        (lambda v: v.day(), lambda vt: vt.day),
        (lambda v: v.hour(), lambda vt: vt.hour),
        (lambda v: v.minute(), lambda vt: vt.minute),
        (lambda v: v.second(), lambda vt: vt.second),
        (lambda v: v.millisecond(), lambda vt: int(vt.microsecond / 1e3)),
    ] + [
        (methodcaller('strftime', pattern), methodcaller('strftime', pattern))
        for pattern in [
            '%Y%m%d %H',
            'DD BAR %w FOO "DD"',
            'DD BAR %w FOO "D',
            'DD BAR "%w" FOO "D',
            'DD BAR "%d" FOO "D',
            'DD BAR "%c" FOO "D',
            'DD BAR "%x" FOO "D',
            'DD BAR "%X" FOO "D',
        ]
    ]
)
def test_timestamp_functions(case_func, expected_func):
    v = L('2015-09-01 14:48:05.359').cast('timestamp')
    vt = datetime.datetime(
        year=2015, month=9, day=1,
        hour=14, minute=48, second=5, microsecond=359000
    )
    result = case_func(v)
    expected = expected_func(vt)
    assert ibis.pandas.execute(result) == expected


@pytest.mark.parametrize(
    'column',
    [
        'datetime_strings_naive',
        'datetime_strings_ny',
        'datetime_strings_utc',
    ]
)
def test_cast_datetime_strings_to_date(t, df, column):
    expr = t[column].cast('date')
    result = expr.execute()
    expected = pd.to_datetime(
        df[column], infer_datetime_format=True
    ).dt.normalize()
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    'column',
    [
        'datetime_strings_naive',
        'datetime_strings_ny',
        'datetime_strings_utc',
    ]
)
def test_cast_datetime_strings_to_timestamp(t, df, column):
    expr = t[column].cast('timestamp')
    result = expr.execute()
    expected = pd.to_datetime(df[column], infer_datetime_format=True)
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    'column',
    [
        'plain_datetimes_naive',
        'plain_datetimes_ny',
        'plain_datetimes_utc',
    ]
)
def test_cast_integer_to_temporal_type(t, df, column):
    expr = t.plain_int64.cast(t[column].type())
    result = expr.execute()
    expected = pd.Series(
        pd.to_datetime(df.plain_int64.values, unit='ns').values,
        index=df.index,
        name='plain_int64',
    ).dt.tz_localize(t[column].type().timezone)
    tm.assert_series_equal(result, expected)


def test_cast_integer_to_date(t, df):
    expr = t.plain_int64.cast('date')
    result = expr.execute()
    expected = pd.Series(
        pd.to_datetime(df.plain_int64.values, unit='D').values,
        index=df.index,
        name='plain_int64',
    )
    tm.assert_series_equal(result, expected)


@pytest.mark.skipif(PY2, reason="not enabled on PY2")
def test_times_ops(t, df):
    result = t.plain_datetimes_naive.time().between('10:00', '10:00').execute()
    expected = np.zeros(len(df), dtype=bool)
    tm.assert_numpy_array_equal(result, expected)

    result = t.plain_datetimes_naive.time().between('01:00', '02:00').execute()
    expected = np.ones(len(df), dtype=bool)
    tm.assert_numpy_array_equal(result, expected)


@pytest.mark.parametrize(
    'tz, rconstruct',
    [
        ('US/Eastern', np.zeros),
        ('UTC', np.ones),
        (None, np.ones)
    ]
)
@pytest.mark.parametrize(
    'column',
    [
        'plain_datetimes_utc',
        'plain_datetimes_naive'
    ]
)
@pytest.mark.skipif(PY2, reason="not enabled on PY2")
def test_times_ops_with_tz(t, df, tz, rconstruct, column):
    expected = rconstruct(len(df), dtype=bool)

    expr = t[column].time().between('01:00', '02:00', timezone=tz)
    result = expr.execute()
    tm.assert_numpy_array_equal(result, expected)

    # Test that casting behavior is the same as using the timezone kwarg
    ts = t[column].cast(dt.Timestamp(timezone=tz))
    expr = ts.time().between('01:00', '02:00')
    result = expr.execute()
    tm.assert_numpy_array_equal(result, expected)


@pytest.mark.skipif(not PY2, reason="testing for PY2")
def test_times_ops_py2(t, df):
    with pytest.raises(ValueError):
        t.plain_datetimes_naive.time()



import ibis
import ibis.expr.datatypes as dt


def test_timestamp_accepts_date_literals(alltypes):
    date_string = '2009-03-01'
    param = ibis.param(dt.timestamp).name('param_0')
    expr = alltypes.mutate(param=param)
    params = {param: date_string}
    result = expr.compile(params=params)
    expected = """\
SELECT *, @param AS `param`
FROM testing.functional_alltypes"""
    assert result == expected



import pytest

from pandas.util import testing as tm
import ibis
from ibis.file.csv import CSVClient, CSVTable
from ibis.file.client import FileDatabase, execute


@pytest.fixture
def transformed(csv):

    # we need to cast to a timestamp type
    # as we read in as strings
    closes = csv.csv_dir.close
    closes = closes.mutate(time=closes.time.cast('timestamp'))
    opens = csv.csv_dir.open
    opens = opens.mutate(time=opens.time.cast('timestamp'))

    t = opens.inner_join(closes, ['time', 'ticker'])
    t = t[opens, closes.close]
    t = t.mutate(avg=(t.open + t.close) / 2)
    t = t[['time', 'ticker', 'avg']]
    return t


def test_client(tmpdir, data):

    # construct with a path to a file
    csv = tmpdir

    for k, v in data.items():
        f = csv / '{}.csv'.format(k)
        v.to_csv(str(f), index=False)

    c = CSVClient(csv / 'open.csv')
    assert c.list_databases() == []
    assert c.list_tables() == ['open']

    c = CSVClient(csv / 'close.csv')
    assert c.list_databases() == []
    assert c.list_tables() == ['close']


def test_navigation(csv):

    # directory navigation
    assert isinstance(csv, FileDatabase)
    result = dir(csv)
    assert result == ['csv_dir']

    prices = csv.csv_dir
    assert isinstance(prices, FileDatabase)
    result = dir(prices)
    assert result == ['close', 'open']
    result = prices.list_tables()
    assert result == ['close', 'open']

    opens = prices.open
    assert isinstance(opens.op(), CSVTable)

    closes = prices.close
    assert isinstance(closes.op(), CSVTable)


def test_read(csv, data):

    closes = csv.csv_dir.close
    assert str(closes) is not None

    result = closes.execute()
    expected = data['close']

    # csv's don't preserve dtypes
    expected['time'] = expected['time'].astype(str)
    tm.assert_frame_equal(result, expected)

    result = execute(closes)
    tm.assert_frame_equal(result, expected)


def test_read_with_projection(csv2, data):

    t = csv2.csv_dir2.df
    result = execute(t)
    assert 'close' in result.columns
    assert 'open' in result.columns

    t = t[['time', 'ticker', 'close']]
    result = execute(t)
    assert 'close' in result.columns
    assert 'open' not in result.columns


def test_insert(transformed, tmpdir):
    t = transformed

    # csv's don't preserve dtypes
    expected = execute(t)
    expected['time'] = expected['time'].astype(str)

    tpath = tmpdir / 'new_csv'
    tpath.mkdir()
    path = tpath / 'foo.csv'

    assert not path.exists()
    c = ibis.csv.connect(tpath)
    c.insert('foo.csv', t)
    assert path.exists()

    # readback
    t = CSVClient(str(tpath)).database()
    result = t.list_tables()
    assert result == ['foo']

    result = t.foo.execute()
    tm.assert_frame_equal(result, expected)



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from posixpath import join as pjoin

import ibis.util as util
import ibis.common as com
import ibis.expr.schema as sch
from ibis.config import options


class DataFrameWriter(object):

    """
    Interface class for writing pandas objects to Impala tables

    Class takes ownership of any temporary data written to HDFS
    """
    def __init__(self, client, df, path=None):
        self.client = client
        self.hdfs = client.hdfs

        self.df = df

        self.temp_hdfs_dirs = []

    def write_temp_csv(self):
        temp_hdfs_dir = pjoin(options.impala.temp_hdfs_path,
                              'pandas_{0}'.format(util.guid()))
        self.hdfs.mkdir(temp_hdfs_dir)

        # Keep track of the temporary HDFS file
        self.temp_hdfs_dirs.append(temp_hdfs_dir)

        # Write the file to HDFS
        hdfs_path = pjoin(temp_hdfs_dir, '0.csv')

        self.write_csv(hdfs_path)

        return temp_hdfs_dir

    def write_csv(self, path):
        import csv

        tmp_path = 'tmp_{0}.csv'.format(util.guid())
        f = open(tmp_path, 'w+')

        try:
            # Write the DataFrame to the temporary file path
            if options.verbose:
                util.log('Writing DataFrame to temporary file')

            self.df.to_csv(f, header=False, index=False,
                           sep=',',
                           quoting=csv.QUOTE_NONE,
                           escapechar='\\',
                           na_rep='#NULL')
            f.seek(0)

            if options.verbose:
                util.log('Writing CSV to: {0}'.format(path))

            self.hdfs.put(path, f)
        finally:
            f.close()
            try:
                os.remove(tmp_path)
            except os.error:
                pass

        return path

    def get_schema(self):
        # define a temporary table using delimited data
        return sch.infer(self.df)

    def delimited_table(self, csv_dir, name=None, database=None):
        temp_delimited_name = 'ibis_tmp_pandas_{0}'.format(util.guid())
        schema = self.get_schema()

        return self.client.delimited_file(csv_dir, schema,
                                          name=temp_delimited_name,
                                          database=database,
                                          delimiter=',',
                                          na_rep='#NULL',
                                          escapechar='\\\\',
                                          external=True,
                                          persist=False)

    def __del__(self):
        try:
            self.cleanup()
        except com.IbisError:
            pass

    def cleanup(self):
        for path in self.temp_hdfs_dirs:
            self.hdfs.rmdir(path)
        self.temp_hdfs_dirs = []
        self.csv_dir = None


def write_temp_dataframe(client, df):
    writer = DataFrameWriter(client, df)
    path = writer.write_temp_csv()
    return writer, writer.delimited_table(path)



from __future__ import absolute_import

import collections

import itertools

import six

import numpy as np
import pandas as pd

from pandas.core.groupby import SeriesGroupBy

import toolz

import ibis.expr.datatypes as dt
import ibis.expr.operations as ops

from ibis.pandas.core import scalar_types
from ibis.pandas.dispatch import execute_node, Dispatcher, pause_ordering
from ibis.compat import (
    functools, signature, Parameter, _empty as empty, viewkeys)


rule_to_python_type = Dispatcher(
    'rule_to_python_type',
    doc="""\
Convert an ibis :class:`~ibis.expr.datatypes.DataType` into a pandas backend
friendly ``multipledispatch`` signature.

Parameters
----------
rule : DataType
    The :class:`~ibis.expr.datatypes.DataType` subclass to map to a pandas
    friendly type.

Returns
-------
types : Union[type, Tuple[type...]]
    A pandas backend friendly signature
"""
)


def arguments_from_signature(signature, *args, **kwargs):
    """Validate signature against `args` and `kwargs` and return the kwargs
    asked for in the signature

    Parameters
    ----------
    args : Tuple[object...]
    kwargs : Dict[str, object]

    Returns
    -------
    args, kwargs : Tuple[Tuple[object...], Dict[str, object]]

    Examples
    --------
    >>> from ibis.compat import signature
    >>> def foo(a, b=1):
    ...     return a + b
    >>> foo_sig = signature(foo)
    >>> args, kwargs = arguments_from_signature(foo_sig, 1, b=2)
    >>> args
    (1,)
    >>> kwargs
    {'b': 2}
    >>> def bar(a):
    ...     return a + 1
    >>> bar_sig = signature(bar)
    >>> args, kwargs = arguments_from_signature(bar_sig, 1, b=2)
    >>> args
    (1,)
    >>> kwargs
    {}
    """
    bound = signature.bind_partial(*args)
    meta_kwargs = toolz.merge({'kwargs': kwargs}, kwargs)
    remaining_parameters = viewkeys(signature.parameters) - viewkeys(
        bound.arguments
    )
    new_kwargs = {
        k: meta_kwargs[k] for k in remaining_parameters
        if k in signature.parameters
        if signature.parameters[k].kind in {
            Parameter.KEYWORD_ONLY,
            Parameter.POSITIONAL_OR_KEYWORD,
            Parameter.VAR_KEYWORD,
        }
    }
    return args, new_kwargs


@rule_to_python_type.register(dt.DataType)
def datatype_rule(rule):
    return scalar_types


@rule_to_python_type.register(dt.Array)
def array_rule(rule):
    return (list,)


@rule_to_python_type.register(dt.Map)
def map_rule(rule):
    return (dict,)


@rule_to_python_type.register(dt.Struct)
def struct_rule(rule):
    return (collections.OrderedDict,)


@rule_to_python_type.register(dt.String)
def string_rule(rule):
    return six.string_types


@rule_to_python_type.register(dt.Integer)
def int_rule(rule):
    return six.integer_types + (np.integer,)


@rule_to_python_type.register(dt.Floating)
def float_rule(rule):
    return (float, np.floating)


def nullable(datatype):
    """Return the signature of a scalar value that is allowed to be NULL (in
    SQL parlance).

    Parameters
    ----------
    datatype : ibis.expr.datatypes.DataType

    Returns
    -------
    Tuple[Type]
    """
    return (type(None),) if datatype.nullable else ()


def udf_signature(input_type, klass):
    """Compute the appropriate signature for an ibis Node from a list of input
    types.

    Parameters
    ----------
    input_type : List[ibis.expr.datatypes.DataType]
        A list of ibis DataType instances
    klass : Type
        pd.Series or SeriesGroupBy

    Returns
    -------
    Tuple[Type]
        A tuple of types appropriate for use in a multiple dispatch signature.

    Examples
    --------
    >>> from pprint import pprint
    >>> import pandas as pd
    >>> from pandas.core.groupby import SeriesGroupBy
    >>> import ibis.expr.datatypes as dt
    >>> input_type = [dt.int64, dt.double]
    >>> sig = udf_signature(input_type, pd.Series)
    >>> pprint(sig)  # doctest: +ELLIPSIS
    ((<class 'pandas.core.series.Series'>,
      <... 'int'>,
      <... 'numpy.integer'>,
      <... 'NoneType'>),
     (<class 'pandas.core.series.Series'>,
      <... 'float'>,
      <... 'numpy.floating'>,
      <... 'NoneType'>))
    >>> input_type = [dt.Int64(nullable=False), dt.Double(nullable=False)]
    >>> sig = udf_signature(input_type, SeriesGroupBy)
    >>> pprint(sig)
    ((<class 'pandas.core.groupby.SeriesGroupBy'>,
      <... 'int'>,
      <... 'numpy.integer'>),
     (<class 'pandas.core.groupby.SeriesGroupBy'>,
      <... 'float'>,
      <... 'numpy.floating'>))
    """
    return tuple(
        (klass,) + rule_to_python_type(r) + nullable(r) for r in input_type
    )


def check_matching_signature(input_type):
    """Make sure that the number of arguments declared by the user in
    `input_type` matches that of the wrapped function's signature.

    Parameters
    ----------
    input_type : List[DataType]

    Returns
    -------
    wrapper : callable
    """
    def wrapper(func):
        num_params = sum(
            param.kind in {param.POSITIONAL_OR_KEYWORD, param.POSITIONAL_ONLY}
            for param in signature(func).parameters.values()
            if param.default is empty
        )
        num_declared = len(input_type)
        if num_params != num_declared:
            raise TypeError(
                'Function {!r} has {:d} parameters, '
                'input_type has {:d}. These must match'.format(
                    func.__name__,
                    num_params,
                    num_declared,
                )
            )

        return func
    return wrapper


def udf(input_type, output_type):
    """Define a UDF (user-defined function) that operates element wise on a
    Pandas Series.

    Parameters
    ----------
    input_type : List[ibis.rules]
        A rules encoding the abstract type of each argument. These are found in
        :mod:`~ibis.expr.rules`. The length of this list must match the number
        of arguments to the function. Argument splatting in the UDF signature
        is not supported.
    output_type : ibis.rules or ibis.expr.datatypes.DataType
        The return type of the function.

    Examples
    --------
    >>> import ibis
    >>> @udf(input_type=[dt.string], output_type=dt.int64)
    ... def my_string_length(series):
    ...     return series.str.len() * 2
    """
    def wrapper(func):
        # generate a new custom node

        UDFNode = type(
            func.__name__,
            (ops.ValueOp,),
            dict(input_type=input_type, output_type=output_type.array_type)
        )

        # Don't reorder the multiple dispatch graph for each of these
        # definitions
        with pause_ordering():

            # Define a execution rule for a simple elementwise Series function
            @execute_node.register(
                UDFNode, *udf_signature(input_type, klass=pd.Series)
            )
            def execute_udf_node(op, *args, **kwargs):
                args, kwargs = arguments_from_signature(
                    signature(func), *args, **kwargs
                )
                return func(*args, **kwargs)

            # Define an execution rule for elementwise operations on a grouped
            # Series
            @execute_node.register(
                UDFNode, *udf_signature(input_type, klass=SeriesGroupBy)
            )
            def execute_udf_node_groupby(op, *args, **kwargs):
                groupers = [
                    grouper for grouper in (
                        getattr(arg, 'grouper', None) for arg in args
                    ) if grouper is not None
                ]

                # all grouping keys must be identical
                assert all(groupers[0] == grouper for grouper in groupers[1:])

                # we're performing a scalar operation on grouped column, so
                # perform the operation directly on the underlying Series and
                # regroup after it's finished
                arguments = [getattr(arg, 'obj', arg) for arg in args]
                groupings = groupers[0].groupings
                args, kwargs = arguments_from_signature(
                    signature(func), *arguments, **kwargs
                )
                return func(*args, **kwargs).groupby(groupings)

        @check_matching_signature(input_type)
        @functools.wraps(func)
        def wrapped(*args):
            return UDFNode(*args).to_expr()

        return wrapped

    return wrapper


def udaf(input_type, output_type):
    """Define a UDAF (user-defined aggregation function) that takes N pandas.
    Series or scalar values as inputs.

    Parameters
    ----------
    input_type : List[T]
        A rules encoding the abstract type of each argument or one of the types
        found in :mod:`~ibis.expr.datatypes`. These are found in
        :mod:`~ibis.expr.rules`. The length of this list must match the number
        of arguments to the function. Argument splatting in the UDF signature
        is not supported.
    output_type : rule or ibis.expr.datatypes.DataType
        The abstract return type of the function. This *cannot* be a rule that
        encodes a :class:`~ibis.expr.types.ColumnExpr` since this API is for
        defining *aggregation* functions.

    Examples
    --------
    >>> import ibis
    >>> @udaf(input_type=[dt.string], output_type=dt.int64)
    ... def my_string_length_agg(series, **kwargs):
    ...     return (series.str.len() * 2).sum()
    """
    def wrapper(func):

        UDAFNode = type(
            func.__name__,
            (ops.Reduction,),
            dict(input_type=input_type, output_type=output_type.scalar_type)
        )

        with pause_ordering():

            # An execution rule for a simple aggregate node
            @execute_node.register(
                UDAFNode, *udf_signature(input_type, klass=pd.Series)
            )
            def execute_udaf_node(op, *args, **kwargs):
                args, kwargs = arguments_from_signature(
                    signature(func), *args, **kwargs
                )
                return func(*args, **kwargs)

            # An execution rule for a grouped aggregation node. This includes
            # aggregates applied over a window.
            @execute_node.register(
                UDAFNode, *udf_signature(input_type, klass=SeriesGroupBy)
            )
            def execute_udaf_node_groupby(op, *args, **kwargs):
                # construct a generator that yields the next group of data for
                # every argument excluding the first (pandas performs the
                # iteration for the first argument) for each argument that is a
                # SeriesGroupBy.
                #
                # If the argument is not a SeriesGroupBy then keep repeating it
                # until all groups are exhausted.
                context = kwargs.pop('context', None)
                assert context is not None, 'context is None'
                iters = (
                    (data for _, data in arg)
                    if isinstance(arg, SeriesGroupBy)
                    else itertools.repeat(arg) for arg in args[1:]
                )
                funcsig = signature(func)

                def aggregator(first, *rest, **kwargs):
                    # map(next, *rest) gets the inputs for the next group
                    # TODO: might be inefficient to do this on every call
                    args, kwargs = arguments_from_signature(
                        funcsig, first, *map(next, rest), **kwargs
                    )
                    return func(*args, **kwargs)

                result = context.agg(args[0], aggregator, *iters, **kwargs)
                return result

        @check_matching_signature(input_type)
        @functools.wraps(func)
        def wrapped(*args):
            return UDAFNode(*args).to_expr()

        return wrapped

    return wrapper



import ibis
from ibis.expr import datatypes as dt


def test_whole_schema():
    customers = ibis.table(
        [
            ('cid', 'int64'),
            ('mktsegment', 'string'),
            ('address', ('struct<city: string, street: string, '
                         'street_number: int32, zip: int16>')),
            ('phone_numbers', 'array<string>'),
            (
                'orders', """array<struct<
                                oid: int64,
                                status: string,
                                totalprice: decimal(12, 2),
                                order_date: string,
                                items: array<struct<
                                    iid: int64,
                                    name: string,
                                    price: decimal(12, 2),
                                    discount_perc: decimal(12, 2),
                                    shipdate: string
                                >>
                            >>"""
            ),
            ('web_visits', ('map<string, struct<user_agent: string, '
                            'client_ip: string, visit_date: string, '
                            'duration_ms: int32>>')),
            ('support_calls', ('array<struct<agent_id: int64, '
                               'call_date: string, duration_ms: int64, '
                               'issue_resolved: boolean, '
                               'agent_comment: string>>'))
        ],
        name='customers',
    )
    expected = ibis.Schema.from_tuples(
        [
            ('cid', dt.int64),
            ('mktsegment', dt.string),
            (
                'address',
                dt.Struct.from_tuples([
                    ('city', dt.string),
                    ('street', dt.string),
                    ('street_number', dt.int32),
                    ('zip', dt.int16)
                ]),
            ),
            ('phone_numbers', dt.Array(dt.string)),
            (
                'orders', dt.Array(dt.Struct.from_tuples([
                            ('oid', dt.int64),
                            ('status', dt.string),
                            ('totalprice', dt.Decimal(12, 2)),
                            ('order_date', dt.string),
                            (
                                'items',
                                dt.Array(dt.Struct.from_tuples([
                                    ('iid', dt.int64),
                                    ('name', dt.string),
                                    ('price', dt.Decimal(12, 2)),
                                    ('discount_perc', dt.Decimal(12, 2)),
                                    ('shipdate', dt.string),
                                ]))
                            )
                        ]))
            ),
            (
                'web_visits',
                dt.Map(
                    dt.string,
                    dt.Struct.from_tuples([
                        ('user_agent', dt.string),
                        ('client_ip', dt.string),
                        ('visit_date', dt.string),
                        ('duration_ms', dt.int32),
                    ])
                )
            ),
            (
                'support_calls',
                dt.Array(dt.Struct.from_tuples([
                    ('agent_id', dt.int64),
                    ('call_date', dt.string),
                    ('duration_ms', dt.int64),
                    ('issue_resolved', dt.boolean),
                    ('agent_comment', dt.string)
                ]))
            ),
        ],
    )
    assert customers.schema() == expected
